"""Correlation charts."""

import vizro.models as vm
import vizro.plotly.express as px

from pages._factories import column_and_line_factory, connected_scatter_factory
from pages._pages_utils import PAGE_GRID, gapminder, iris, make_code_clipboard_from_py_file

scatter = vm.Page(
    title="Scatter",
    path="correlation/scatter",
    layout=vm.Layout(grid=PAGE_GRID),
    components=[
        vm.Card(
            text="""

            #### What is a scatter chart?

            A scatter plot is a two-dimensional data visualization using dots to represent the values obtained for two
            different variables - one plotted along the x-axis and the other plotted along the y-axis.

            &nbsp;

            #### When should I use it?

            Use scatter plots when you want to show the relationship between two variables. Scatter plots are sometimes
            called _Correlation plots_ because they show how two variables are correlated. Scatter plots are ideal when
            you have paired numerical data and you want to see if one variable impacts the other. However, do remember
            that correlation is not causation. Make sure your audience does not draw the wrong conclusions.
        """
        ),
        vm.Graph(figure=px.scatter(iris, x="sepal_width", y="sepal_length", color="species")),
        make_code_clipboard_from_py_file("scatter.py"),
    ],
)

connected_scatter = connected_scatter_factory("correlation")

scatter_matrix = vm.Page(
    title="Scatter matrix",
    path="correlation/scatter-matrix",
    layout=vm.Layout(grid=PAGE_GRID),
    components=[
        vm.Card(
            text="""

            #### What is a scatter matrix?

            A scatter matrix, also known as a SPLOM chart, is a multi-dimensional data visualization that displays
            scatter plots for every pair of variables in a dataset. Each scatter plot is positioned in a matrix format,
            where rows and columns represent different variables.

            &nbsp;

            #### When should I use it?

            Use a scatter matrix when you want to explore relationships between multiple pairs of variables
            simultaneously. They are particularly useful for identifying correlations, patterns, and potential outliers
            within a dataset containing multiple numerical variables. Carefully select the most relevant variables to
            ensure clarity and readability of the chart.
        """
        ),
        vm.Graph(
            figure=px.scatter_matrix(iris, dimensions=["sepal_length", "sepal_width", "petal_length", "petal_width"])
        ),
        make_code_clipboard_from_py_file("scatter_matrix.py"),
    ],
)

bubble = vm.Page(
    title="Bubble",
    path="correlation/bubble",
    layout=vm.Layout(grid=PAGE_GRID),
    components=[
        vm.Card(
            text="""

            #### What is a bubble chart?

            A bubble chart is a type of data visualization that displays three dimensions of data. Each point on the
            chart is represented by a bubble, where the x-axis and y-axis denote two of the data dimensions, and the
            size of the bubble represents the third dimension.

            &nbsp;

            #### When should I use it?

            Use a bubble chart when you want to explore and compare relationships between three variables
            simultaneously. They are particularly useful for identifying patterns, trends, and outliers in
            multi-dimensional data. Bubble charts can help you visualize the impact of a third variable,
            providing deeper insights than a standard scatter plot.
        """
        ),
        vm.Graph(figure=px.scatter(gapminder.query("year==2007"), x="gdpPercap", y="lifeExp", size="pop", size_max=60)),
        make_code_clipboard_from_py_file("bubble.py"),
    ],
)

column_and_line = column_and_line_factory("correlation")
pages = [scatter, connected_scatter, scatter_matrix, bubble, column_and_line]



"""Contains reusable page functions to create identical content with a different `id`.

Note: Since each page can only belong to one navigation group, we need a new page with a unique ID for
each chart type used in different groups.
"""

import vizro.models as vm
import vizro.plotly.express as px
from custom_charts import butterfly, column_and_line

from pages._pages_utils import PAGE_GRID, ages, gapminder, make_code_clipboard_from_py_file


def butterfly_factory(group: str):
    """Reusable function to create the page content for the butterfly chart with a unique ID."""
    return vm.Page(
        id=f"{group}-butterfly",
        path=f"{group}/butterfly",
        title="Butterfly",
        layout=vm.Layout(grid=PAGE_GRID),
        components=[
            vm.Card(
                text="""

                #### What is a butterfly chart?

                A butterfly chart (also called a tornado chart) is a bar chart for displaying two sets of data series
                side by side.

                &nbsp;

                #### When should I use it?

                Use a butterfly chart when you wish to emphasize the comparison between two data sets sharing the same
                parameters. Sharing this chart with your audience will help them see at a glance how two groups differ
                within the same parameters. You can also **stack** two bars on each side if you wish to divide your
                categories.
            """
            ),
            vm.Graph(figure=butterfly(ages, x1="Male", x2="Female", y="Age")),
            make_code_clipboard_from_py_file("butterfly.py"),
        ],
    )


def connected_scatter_factory(group: str):
    """Reusable function to create the page content for the column chart with a unique ID."""
    return vm.Page(
        id=f"{group}-connected-scatter",
        path=f"{group}/connected-scatter",
        title="Connected scatter",
        layout=vm.Layout(grid=PAGE_GRID),
        components=[
            vm.Card(
                text="""
                #### What is a connected scatter chart?

                A connected scatter chart visualizes two variables (x and y) using dots, with lines connecting the dots
                in the order of the data points. One variable is plotted along the x-axis and the other along the
                y-axis, showing both the relationship and a sequence of the data.

                &nbsp;

                #### When should I use it?

                Use connected scatter charts to show the relationship between two variables and the sequence of data
                points. They are ideal for paired numerical data, helping to reveal trends and patterns over time or in
                a specific order. Remember, correlation is not causation, so ensure your audience understands this to
                avoid misinterpretation.
        """
            ),
            vm.Graph(figure=px.line(gapminder.query("country == 'Australia'"), x="year", y="lifeExp", markers=True)),
            make_code_clipboard_from_py_file("connected_scatter.py"),
        ],
    )


def column_and_line_factory(group: str):
    """Reusable function to create the page content for the column+line chart with a unique ID."""
    return vm.Page(
        id=f"{group}-column-and-line",
        path=f"{group}/column-and-line",
        title="Column and line",
        layout=vm.Layout(grid=PAGE_GRID),
        components=[
            vm.Card(
                text="""
                #### What is a column and line chart?

                A combined column and line chart helps you demonstrate the relationship between an amount
                (displayed in columns) and a trend or rate (displayed as a line running across the columns).

                &nbsp;

                #### When should I use it?

                Use this type of chart when you wish to compare quantities of one item with changes in another item.
                It's ideal for showing patterns over time (e.g., monthly sales and growth rates) but can also be used
                for other types of data comparisons.
        """
            ),
            vm.Graph(
                figure=column_and_line(
                    gapminder.query("country == 'Vietnam'"),
                    y_column="gdpPercap",
                    y_line="lifeExp",
                    x="year",
                )
            ),
            make_code_clipboard_from_py_file("column_and_line.py"),
        ],
    )



from plotly import graph_objects as go


class _DashboardReadyFigure(go.Figure):
    # Just for IDE completion and to define new attribute
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        data = args[0] if args else None
        #  refers to the data argument of https://plotly.com/python-api-reference/generated/plotly.graph_objects.Figure.html
        if isinstance(data, dict):
            self._captured_callable = data.get("_captured_callable", None)

    # This fixes copy.deepcopy(_DashboardReadyFigure), but pickling fails in case of functions decorated with `capture`
    # https://stackoverflow.com/questions/52185507/pickle-and-decorated-classes-picklingerror-not-the-same-object
    # If wanting to save _DashboardReadyFigure objects, we currently recommend "dill" (see also example in link)
    # https://pypi.org/project/dill/
    def __reduce__(self):
        _, (props,) = super().__reduce__()
        props["_captured_callable"] = self._captured_callable
        return self.__class__, (props,)



"""Contains default KPI card functions."""

from typing import Optional

import dash_bootstrap_components as dbc
import numpy as np
import pandas as pd
from dash import html

from vizro.models.types import capture


@capture("figure")
def kpi_card(  # noqa: PLR0913
    data_frame: pd.DataFrame,
    value_column: str,
    *,
    value_format: str = "{value}",
    agg_func: str = "sum",
    title: Optional[str] = None,
    icon: Optional[str] = None,
) -> dbc.Card:
    """Creates a styled KPI (Key Performance Indicator) card displaying a value.

    **Warning:** Note that the format string provided to `value_format` is being evaluated, so ensure that only trusted
    user input is provided to prevent
    [potential security risks](https://stackoverflow.com/questions/76783239/is-it-safe-to-use-python-str-format-method-with-user-submitted-templates-in-serv).

    Args:
        data_frame: DataFrame containing the data.
        value_column: Column name of the value to be shown.
        value_format: Format string to be applied to the value. It must be a
            [valid Python format](https://docs.python.org/3/library/string.html#format-specification-mini-language)
            string where any of the below placeholders can be used. Defaults to "{value}".

            - value: `value_column` aggregated by `agg_func`.

            **Common examples include:**

             - "{value}": Displays the raw value.
             - "${value:0.2f}": Formats the value as a currency with two decimal places.
             - "{value:.0%}": Formats the value as a percentage without decimal places.
             - "{value:,}": Formats the value with comma as a thousands separator.

        agg_func: String function name to be used for aggregating the data. Common options include
            "sum", "mean" or "median". Default is "sum". For more information on possible functions, see
            https://stackoverflow.com/questions/65877567/passing-function-names-as-strings-to-pandas-groupby-aggregrate.
        title: KPI title displayed on top of the card. If not provided, it defaults to the capitalized `value_column`.
        icon: Name of the icon from the [Google Material Icon Library](https://fonts.google.com/icons) to be displayed
            on the left side of the KPI title. If not provided, no icon is displayed.

    Raises:
        UserWarning: If `value_format` is provided, a warning is raised to make aware that only trusted user
            input should be provided.

    Returns:
         A Dash Bootstrap Components card (`dbc.Card`) containing the formatted KPI value.

    Examples:
        Wrap inside `vm.Figure` to use as a component inside `vm.Page` or `vm.Container`.
        >>> import vizro.models as vm
        >>> from vizro.figures import kpi_card
        >>> vm.Page(title="Page", components=[vm.Figure(figure=kpi_card(...))])

    """
    title = title or f"{agg_func} {value_column}".title()
    value = data_frame[value_column].agg(agg_func)

    header = dbc.CardHeader(
        [
            html.P(icon, className="material-symbols-outlined") if icon else None,
            html.H4(title, className="card-kpi-title"),
        ]
    )
    body = dbc.CardBody(value_format.format(value=value))
    return dbc.Card([header, body], className="card-kpi")


@capture("figure")
def kpi_card_reference(  # noqa: PLR0913
    data_frame: pd.DataFrame,
    value_column: str,
    reference_column: str,
    *,
    value_format: str = "{value}",
    reference_format: str = "{delta_relative:+.1%} vs. reference ({reference})",
    agg_func: str = "sum",
    title: Optional[str] = None,
    icon: Optional[str] = None,
) -> dbc.Card:
    """Creates a styled KPI (Key Performance Indicator) card displaying a value in comparison to a reference value.

    **Warning:** Note that the format string provided to `value_format` and `reference_format` is being evaluated,
    so ensure that only trusted user input is provided to prevent
    [potential security risks](https://stackoverflow.com/questions/76783239/is-it-safe-to-use-python-str-format-method-with-user-submitted-templates-in-serv).

    Args:
        data_frame: DataFrame containing the data.
        value_column: Column name of the value to be shown.
        reference_column: Column name of the reference value for comparison.
        value_format: Format string to be applied to the value. It must be a
            [valid Python format](https://docs.python.org/3/library/string.html#format-specification-mini-language)
            string where any of the below placeholders can be used. Defaults to "{value}".

            - value: `value_column` aggregated by `agg_func`.
            - reference: `reference_column` aggregated by `agg_func`.
            - delta: Difference between `value` and `reference`.
            - delta_relative: Relative difference between `value` and `reference`.

            **Common examples include:**

             - "{value}": Displays the raw value.
             - "${value:0.2f}": Formats the value as a currency with two decimal places.
             - "{value:.0%}": Formats the value as a percentage without decimal places.
             - "{value:,}": Formats the value with comma as a thousands separator.

        reference_format: Format string to be applied to the reference. For more details on possible placeholders, see
            docstring on `value_format`. Defaults to "{delta_relative:+.1%} vs. reference ({reference})".
        agg_func: String function name to be used for aggregating the data. Common options include
            "sum", "mean" or "median". Default is "sum". For more information on possible functions, see
            https://stackoverflow.com/questions/65877567/passing-function-names-as-strings-to-pandas-groupby-aggregrate.
        title: KPI title displayed on top of the card. If not provided, it defaults to the capitalized `value_column`.
        icon: Name of the icon from the [Google Material Icon Library](https://fonts.google.com/icons) to be displayed
            on the left side of the KPI title. If not provided, no icon is displayed.

    Raises:
        UserWarning: If `value_format` or `reference_format` is provided, a warning is raised to make aware
            that only trusted user input should be provided.

    Returns:
        A Dash Bootstrap Components card (`dbc.Card`) containing the formatted KPI value and reference.

    Examples:
        Wrap inside `vm.Figure` to use as a component inside `vm.Page` or `vm.Container`.
        >>> import vizro.models as vm
        >>> from vizro.figures import kpi_card_reference
        >>> vm.Page(title="Page", components=[vm.Figure(figure=kpi_card_reference(...))])

    """
    title = title or f"{agg_func} {value_column}".title()
    value, reference = data_frame[[value_column, reference_column]].agg(agg_func)
    delta = value - reference
    delta_relative = delta / reference if reference else np.nan

    header = dbc.CardHeader(
        [
            html.P(icon, className="material-symbols-outlined") if icon else None,
            html.H4(title, className="card-kpi-title"),
        ]
    )
    body = dbc.CardBody(
        value_format.format(value=value, reference=reference, delta=delta, delta_relative=delta_relative)
    )
    footer = dbc.CardFooter(
        [
            html.Span(
                "arrow_circle_up" if delta > 0 else "arrow_circle_down" if delta < 0 else "arrow_circle_right",
                className="material-symbols-outlined",
            ),
            html.Span(
                reference_format.format(value=value, reference=reference, delta=delta, delta_relative=delta_relative)
            ),
        ],
        className="color-pos" if delta > 0 else "color-neg" if delta < 0 else "",
    )
    return dbc.Card([header, body, footer], className="card-kpi")



import dash_ag_grid as dag
import pandas as pd
from asserts import assert_component_equal
from pandas import Timestamp
from vizro.models.types import capture
from vizro.tables import dash_ag_grid

data = pd.DataFrame(
    {
        "cat": ["a", "b", "c"],
        "int": [4, 5, 6],
        "float": [7.3, 8.2, 9.1],
        "date": pd.to_datetime(["2021/01/01", "2021/01/02", "2021/01/03"]),
    }
)
column_defs = [{"field": "cat"}, {"field": "int"}, {"field": "float"}, {"field": "date"}]
row_data_date_converted = [
    {"cat": "a", "int": 4, "float": 7.3, "date": "2021-01-01"},
    {"cat": "b", "int": 5, "float": 8.2, "date": "2021-01-02"},
    {"cat": "c", "int": 6, "float": 9.1, "date": "2021-01-03"},
]
row_data_date_raw = [
    {"cat": "a", "date": Timestamp("2021-01-01 00:00:00"), "float": 7.3, "int": 4},
    {"cat": "b", "date": Timestamp("2021-01-02 00:00:00"), "float": 8.2, "int": 5},
    {"cat": "c", "date": Timestamp("2021-01-03 00:00:00"), "float": 9.1, "int": 6},
]
default_col_defs = {
    "filter": True,
    "filterParams": {"buttons": ["apply", "reset"], "closeOnApply": True},
    "resizable": True,
    "sortable": True,
}
style = {"height": "100%"}


class TestDashAgGrid:
    def test_dash_ag_grid(self):
        grid = dash_ag_grid(data_frame=data)()
        assert_component_equal(
            grid,
            dag.AgGrid(
                columnDefs=column_defs,
                rowData=row_data_date_converted,
                defaultColDef=default_col_defs,
                style=style,
            ),
            keys_to_strip={"className", "dashGridOptions"},
        )
        # skipping only dashGridOptions as this is mostly our defaults for data formats, and would crowd the tests

    def test_custom_dash_ag_grid(self):
        @capture("ag_grid")
        def custom_dash_ag_grid(data_frame):
            return dag.AgGrid(
                columnDefs=[{"field": col} for col in data_frame.columns],
                rowData=data_frame.to_dict("records"),
            )

        ag_grid = custom_dash_ag_grid(data_frame=data)()

        assert_component_equal(
            ag_grid,
            dag.AgGrid(
                rowData=row_data_date_raw,
                columnDefs=column_defs,
            ),
        )



import logging
from functools import wraps

from vizro.models.types import CapturedCallable, _SupportsCapturedCallable

logger = logging.getLogger(__name__)


def _log_call(method):
    @wraps(method)
    def _wrapper(self, *args, **kwargs):
        # We need to run method before logging so that @_log_call works for __init__.
        return_value = method(self, *args, **kwargs)
        logger.debug("Running %s.%s for model with id %s", self.__class__.__name__, method.__name__, self.id)
        return return_value

    return _wrapper


# Validators for reuse
def validate_min_length(cls, value):
    if not value:
        raise ValueError("Ensure this value has at least 1 item.")
    return value


def check_captured_callable(cls, value):
    if isinstance(value, CapturedCallable):
        captured_callable = value
    elif isinstance(value, _SupportsCapturedCallable):
        captured_callable = value._captured_callable
    else:
        return value

    raise ValueError(
        f"A callable of mode `{captured_callable._mode}` has been provided. Please wrap it inside "
        f"`{captured_callable._model_example}`."
    )


REPLACEMENT_STRINGS = {
    # "substring to match a larger general module string": "string to replace with"
    # dot required so that in the case where no replacement is used, we do not
    # have a preceding dot (see __repr_clean__ in types.py)
    "plotly.express": "px.",
    "vizro.tables": "vt.",
    "vizro.figures": "vf.",
    "vizro.actions": "va.",
    "vizro.charts": "vc.",
}



"""Distribution charts."""

import vizro.models as vm
import vizro.plotly.express as px

from pages._factories import butterfly_factory
from pages._pages_utils import PAGE_GRID, make_code_clipboard_from_py_file, tips

violin = vm.Page(
    title="Violin",
    path="distribution/violin",
    layout=vm.Layout(grid=PAGE_GRID),
    components=[
        vm.Card(
            text="""

            #### What is a violin chart?

            A violin chart is similar to a box plot, but works better for visualizing more complex distributions and
            their probability density at different values.

            &nbsp;

            #### When should I use it?

            Use this chart to go beyond the simple box plot and show the distribution shape of the data, the
            inter-quartile range, the confidence intervals and the median.
        """
        ),
        vm.Graph(
            figure=px.violin(
                tips,
                y="total_bill",
                x="day",
                color="day",
                box=True,
            )
        ),
        make_code_clipboard_from_py_file("violin.py"),
    ],
)

boxplot = vm.Page(
    title="Boxplot",
    path="distribution/boxplot",
    layout=vm.Layout(grid=PAGE_GRID),
    components=[
        vm.Card(
            text="""

            #### What is a boxplot?

            A box plot (also known as whisker plot) provides a visual display of multiple datasets,
            indicating the median (center) and the range of the data for each.

            &nbsp;

            #### When should I use it?

            Choose a box plot when you need to summarize distributions between many groups or datasets. It takes up
            less space than many other charts.

            Create boxes to display the median, and the upper and lower quartiles. Add whiskers to highlight
            variability outside the upper and lower quartiles. You can add outliers as dots beyond, but in line with
            the whiskers.
        """
        ),
        vm.Graph(
            figure=px.box(
                tips,
                y="total_bill",
                x="day",
                color="day",
            )
        ),
        make_code_clipboard_from_py_file("boxplot.py"),
    ],
)

butterfly = butterfly_factory("distribution")

histogram = vm.Page(
    title="Histogram",
    path="distribution/histogram",
    layout=vm.Layout(grid=PAGE_GRID),
    components=[
        vm.Card(
            text="""
            #### What is a histogram?

            A histogram organizes numerical data into columns, with the size of each column representing how frequently
            values fall within specified ranges. It visualizes data across a continuous interval.

            &nbsp;

            #### When should I use it?

            A histogram is useful for showing your audience where specific values are concentrated, identifying the
            extremes, and spotting any gaps or outliers. It can also help you visualize a rough probability
            distribution. Ensure that the gaps between columns are minimal to make the 'shape' of your data
            immediately clear.
        """
        ),
        vm.Graph(figure=px.histogram(tips, x="total_bill")),
        make_code_clipboard_from_py_file("histogram.py"),
    ],
)


pages = [violin, boxplot, butterfly, histogram]



import vizro.models as vm
import vizro.plotly.express as px
from vizro import Vizro

gapminder = px.data.gapminder()

page = vm.Page(
    title="Choropleth",
    components=[
        vm.Graph(
            figure=px.choropleth(
                gapminder.query("year == 2007"), locations="iso_alpha", color="lifeExp", hover_name="country"
            )
        )
    ],
)

dashboard = vm.Dashboard(pages=[page])
Vizro().build(dashboard).run()



import logging
import os

import plotly.io as pio
from dash.development.base_component import Component

from ._themes import dark, light
from ._vizro import Vizro

logging.basicConfig(level=os.getenv("VIZRO_LOG_LEVEL", "WARNING"))
pio.templates["vizro_dark"] = dark
pio.templates["vizro_light"] = light

__all__ = ["Vizro"]

__version__ = "0.1.21"


# For the below _css_dist to be used by Dash, it must be retrieved by dash.resources.Css.get_all_css(). This means it
# must be added to dash.development.base_component.ComponentRegistry. The simplest way to do this is to run
# ComponentRegistry.registry.add("vizro") and this appears to be sufficient for our needs, but it is not documented
# anywhere. The same function is run (together with some others which we probably don't need) when subclassing
# Component, thanks to the metaclass ComponentMeta. So we define a dummy component to go through this safer route,
# even though we don't need the component for anything. _css_dist is automatically served on import of vizro, regardless
# of whether the Vizro class or any other bits are used.
class _Dummy(Component):
    pass


# For dev versions, a branch or tag called e.g. 0.1.20.dev0 does not exist and so won't work with the CDN. We point
# to main instead, but this can be manually overridden to the current feature branch name if required.
# _git_branch = __version__ if "dev" not in __version__ else "main"
_git_branch = __version__ if "dev" not in __version__ else "main"
_library_css = ["static/css/figures"]
_base_external_url = f"https://cdn.jsdelivr.net/gh/mckinsey/vizro@{_git_branch}/vizro-core/src/vizro/"

# CSS is packaged and accessed using relative_package_path when serve_locally=False (the default) in
# the Dash instantiation. When serve_locally=True then, where defined, external_url will be used instead.
_css_dist = [
    {
        "namespace": "vizro",
        "relative_package_path": f"{css_file}.css",
        "external_url": f"{_base_external_url}{css_file}.min.css",
    }
    for css_file in _library_css
]

# Include font file so that figures with icons can be used outside Vizro as pure Dash components.
# The file can be served through the CDN in the same way as the CSS files but external_url is irrelevant here. The way
# the file is requested is through a relative url("./fonts/...") in the requesting CSS file. When the CSS file is
# served from the CDN then this will refer to the font file also on the CDN.
_css_dist.append(
    {
        "namespace": "vizro",
        "relative_package_path": "static/css/fonts/material-symbols-outlined.woff2",
    }
)



import vizro.models as vm
import vizro.plotly.express as px
from vizro import Vizro

iris = px.data.iris()

page = vm.Page(
    title="Scatter matrix",
    components=[
        vm.Graph(
            figure=px.scatter_matrix(iris, dimensions=["sepal_length", "sepal_width", "petal_length", "petal_width"])
        )
    ],
)

dashboard = vm.Dashboard(pages=[page])
Vizro().build(dashboard).run()



import vizro.models as vm
import vizro.plotly.express as px
from vizro import Vizro

tips = px.data.tips()

page = vm.Page(
    title="Heatmap",
    components=[
        vm.Graph(figure=px.density_heatmap(tips, x="day", y="size", z="tip", histfunc="avg", text_auto="$.2f")),
    ],
)

dashboard = vm.Dashboard(pages=[page])
Vizro().build(dashboard).run()



import logging
from typing import Dict, List, Literal

import pandas as pd
from dash import State, dcc, html

try:
    from pydantic.v1 import Field, PrivateAttr, validator
except ImportError:  # pragma: no cov
    from pydantic import Field, PrivateAttr, validator

from vizro.actions._actions_utils import CallbackTriggerDict, _get_component_actions, _get_parent_vizro_model
from vizro.managers import data_manager
from vizro.models import Action, VizroBaseModel
from vizro.models._action._actions_chain import _action_validator_factory
from vizro.models._components._components_utils import _process_callable_data_frame
from vizro.models._models_utils import _log_call
from vizro.models.types import CapturedCallable

logger = logging.getLogger(__name__)


class Table(VizroBaseModel):
    """Wrapper for `dash_table.DataTable` to visualize tables in dashboard.

    Args:
        type (Literal["table"]): Defaults to `"table"`.
        figure (CapturedCallable): Function that returns a Dash DataTable. See [`vizro.tables`][vizro.tables].
        title (str): Title of the table. Defaults to `""`.
        actions (List[Action]): See [`Action`][vizro.models.Action]. Defaults to `[]`.

    """

    type: Literal["table"] = "table"
    figure: CapturedCallable = Field(
        ..., import_path="vizro.tables", mode="table", description="Function that returns a `Dash DataTable`."
    )
    title: str = Field("", description="Title of the table")
    actions: List[Action] = []

    _input_component_id: str = PrivateAttr()

    # Component properties for actions and interactions
    _output_component_property: str = PrivateAttr("children")

    # Validators
    set_actions = _action_validator_factory("active_cell")
    _validate_callable = validator("figure", allow_reuse=True, always=True)(_process_callable_data_frame)

    # Convenience wrapper/syntactic sugar.
    def __call__(self, **kwargs):
        kwargs.setdefault("data_frame", data_manager[self["data_frame"]].load())
        figure = self.figure(**kwargs)
        figure.id = self._input_component_id
        return figure

    # Convenience wrapper/syntactic sugar.
    def __getitem__(self, arg_name: str):
        # See figure implementation for more details.
        if arg_name == "type":
            return self.type
        return self.figure[arg_name]

    # Interaction methods
    @property
    def _filter_interaction_input(self):
        """Required properties when using pre-defined `filter_interaction`."""
        return {
            "active_cell": State(component_id=self._input_component_id, component_property="active_cell"),
            "derived_viewport_data": State(
                component_id=self._input_component_id,
                component_property="derived_viewport_data",
            ),
            "modelID": State(component_id=self.id, component_property="id"),  # required, to determine triggered model
        }

    def _filter_interaction(
        self, data_frame: pd.DataFrame, target: str, ctd_filter_interaction: Dict[str, CallbackTriggerDict]
    ) -> pd.DataFrame:
        """Function to be carried out for pre-defined `filter_interaction`."""
        # data_frame is the DF of the target, ie the data to be filtered, hence we cannot get the DF from this model
        ctd_active_cell = ctd_filter_interaction["active_cell"]
        ctd_derived_viewport_data = ctd_filter_interaction["derived_viewport_data"]
        if not ctd_active_cell["value"] or not ctd_derived_viewport_data["value"]:
            return data_frame

        # ctd_active_cell["id"] represents the underlying table id, so we need to fetch its parent Vizro Table actions.
        source_table_actions = _get_component_actions(_get_parent_vizro_model(ctd_active_cell["id"]))

        for action in source_table_actions:
            if action.function._function.__name__ != "filter_interaction" or target not in action.function["targets"]:
                continue
            column = ctd_active_cell["value"]["column_id"]
            derived_viewport_data_row = ctd_active_cell["value"]["row"]
            clicked_data = ctd_derived_viewport_data["value"][derived_viewport_data_row][column]
            data_frame = data_frame[data_frame[column].isin([clicked_data])]

        return data_frame

    @_log_call
    def pre_build(self):
        self._input_component_id = self.figure._arguments.get("id", f"__input_{self.id}")

    def build(self):
        return dcc.Loading(
            children=[
                html.H3(self.title) if self.title else None,
                # Refer to the vm.AgGrid build method for details on why we return the
                # html.Div(id=self._input_component_id) instead of actual figure object with the original data_frame.
                html.Div(
                    id=self.id,
                    children=[html.Div(id=self._input_component_id)],
                    className="table-container",
                ),
            ],
            color="grey",
            parent_className="loading-container",
            overlay_style={"visibility": "visible", "opacity": 0.3},
        )



import pandas as pd
import vizro.models as vm
import vizro.plotly.express as px
from vizro import Vizro

stepped_line_data = pd.DataFrame(
    {
        "year": [1995, 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003],
        "rate": [0.10, 0.12, 0.15, 0.13, 0.14, 0.13, 0.14, 0.16, 0.15],
    }
)

page = vm.Page(
    title="Stepped line",
    components=[
        vm.Graph(
            figure=px.line(
                data_frame=stepped_line_data,
                x="year",
                y="rate",
                line_shape="vh",
            ),
        )
    ],
)
dashboard = vm.Dashboard(pages=[page])
Vizro().build(dashboard).run()



from typing import List, Literal

import dash_bootstrap_components as dbc
from dash import html

try:
    from pydantic.v1 import Field
except ImportError:  # pragma: no cov
    from pydantic import Field

from vizro.models import Action, VizroBaseModel
from vizro.models._action._actions_chain import _action_validator_factory
from vizro.models._models_utils import _log_call


class UserInput(VizroBaseModel):
    """Single-line text input component `UserInput`.

    Based on the underlying [`dcc.Input`](https://dash.plotly.com/dash-core-components/input).

    Args:
        type (Literal["user_input"]): Defaults to `"user_input"`.
        title (str): Title to be displayed. Defaults to `""`.
        placeholder (str): Default text to display in input field. Defaults to `""`.
        actions (Optional[List[Action]]): Defaults to `[]`.

    """

    type: Literal["user_input"] = "user_input"
    # TODO: before making public consider naming this field (or giving an alias) label instead of title
    title: str = Field("", description="Title to be displayed")
    placeholder: str = Field("", description="Default text to display in input field")
    actions: List[Action] = []

    # Re-used validators
    # TODO: Before making public, consider how actions should be triggered and what the default property should be
    # See comment thread: https://github.com/mckinsey/vizro/pull/298#discussion_r1478137654
    _set_actions = _action_validator_factory("value")

    @_log_call
    def build(self):
        return html.Div(
            children=[
                dbc.Label(children=self.title, html_for=self.id) if self.title else None,
                dbc.Input(
                    id=self.id,
                    placeholder=self.placeholder,
                    type="text",
                    persistence=True,
                    persistence_type="session",
                    debounce=True,
                ),
            ],
        )



from typing import List

import pandas as pd
import plotly.graph_objects as go
import vizro.models as vm
from vizro import Vizro
from vizro.models.types import capture

sankey_data = pd.DataFrame(
    {
        "Origin": [0, 1, 0, 2, 3, 3],  # indices inside labels
        "Destination": [2, 3, 3, 4, 4, 5],  # indices inside labels
        "Value": [8, 4, 2, 8, 4, 2],
    }
)


@capture("graph")
def sankey(
    data_frame: pd.DataFrame,
    source: str,
    target: str,
    value: str,
    labels: List[str],
):
    fig = go.Figure(
        data=[
            go.Sankey(
                node={
                    "pad": 16,
                    "thickness": 16,
                    "label": labels,
                },
                link={
                    "source": data_frame[source],
                    "target": data_frame[target],
                    "value": data_frame[value],
                    "label": labels,
                    "color": "rgba(205, 209, 228, 0.4)",
                },
            )
        ]
    )
    fig.update_layout(barmode="relative")
    return fig


page = vm.Page(
    title="Sankey",
    components=[
        vm.Graph(
            figure=sankey(
                sankey_data,
                labels=["A1", "A2", "B1", "B2", "C1", "C2"],
                source="Origin",
                target="Destination",
                value="Value",
            ),
        ),
    ],
)

dashboard = vm.Dashboard(pages=[page])
Vizro().build(dashboard).run()



"""Unit tests for vizro.models.Table."""

import re

import pytest
from asserts import assert_component_equal
from dash import dcc, html

try:
    from pydantic.v1 import ValidationError
except ImportError:  # pragma: no cov
    from pydantic import ValidationError

import vizro.models as vm
import vizro.plotly.express as px
from vizro.managers import data_manager
from vizro.models._action._action import Action
from vizro.tables import dash_data_table


@pytest.fixture
def dash_table_with_arguments():
    return dash_data_table(data_frame=px.data.gapminder(), style_header={"border": "1px solid green"})


@pytest.fixture
def dash_table_with_str_dataframe():
    return dash_data_table(data_frame="gapminder")


class TestTableInstantiation:
    def test_create_graph_mandatory_only(self, standard_dash_table):
        table = vm.Table(figure=standard_dash_table)

        assert hasattr(table, "id")
        assert table.type == "table"
        assert table.figure == standard_dash_table
        assert table.actions == []

    @pytest.mark.parametrize("id", ["id_1", "id_2"])
    def test_create_table_mandatory_and_optional(self, standard_dash_table, id):
        table = vm.Table(id=id, figure=standard_dash_table, actions=[])

        assert table.id == id
        assert table.type == "table"
        assert table.figure == standard_dash_table

    def test_mandatory_figure_missing(self):
        with pytest.raises(ValidationError, match="field required"):
            vm.Table()

    def test_captured_callable_invalid(self, standard_go_chart):
        with pytest.raises(
            ValidationError,
            match=re.escape(
                "Invalid CapturedCallable. Supply a function imported from vizro.tables or "
                "defined with decorator @capture('table')."
            ),
        ):
            vm.Table(figure=standard_go_chart)

    def test_captured_callable_wrong_mode(self, standard_ag_grid):
        with pytest.raises(
            ValidationError,
            match=re.escape(
                "CapturedCallable was defined with @capture('ag_grid') rather than @capture('table') and so "
                "is not compatible with the model."
            ),
        ):
            vm.Table(figure=standard_ag_grid)

    def test_is_model_inheritable(self, standard_dash_table):
        class MyTable(vm.Table):
            pass

        my_table = MyTable(figure=standard_dash_table)

        assert hasattr(my_table, "id")
        assert my_table.type == "table"
        assert my_table.figure == standard_dash_table
        assert my_table.actions == []

    def test_set_action_via_validator(self, standard_dash_table, identity_action_function):
        table = vm.Table(figure=standard_dash_table, actions=[Action(function=identity_action_function())])
        actions_chain = table.actions[0]
        assert actions_chain.trigger.component_property == "active_cell"


class TestDunderMethodsTable:
    def test_getitem_known_args(self, dash_table_with_arguments):
        table = vm.Table(figure=dash_table_with_arguments)
        assert table["style_header"] == {"border": "1px solid green"}
        assert table["type"] == "table"

    def test_getitem_unknown_args(self, standard_dash_table):
        table = vm.Table(figure=standard_dash_table)
        with pytest.raises(KeyError):
            table["unknown_args"]

    def test_underlying_id_is_auto_generated(self, standard_dash_table):
        table = vm.Table(id="table", figure=standard_dash_table)
        table.pre_build()
        # table() is the same as table.__call__()
        assert table().id == "__input_table"

    def test_underlying_id_is_provided(self, dash_data_table_with_id):
        table = vm.Table(figure=dash_data_table_with_id)
        table.pre_build()
        # table() is the same as table.__call__()
        assert table().id == "underlying_table_id"


class TestAttributesTable:
    def test_table_filter_interaction_attributes(self, dash_data_table_with_id):
        table = vm.Table(figure=dash_data_table_with_id, title="Gapminder", actions=[])
        table.pre_build()
        assert hasattr(table, "_filter_interaction_input")
        assert "modelID" in table._filter_interaction_input


class TestProcessTableDataFrame:
    # Testing at this low implementation level as mocking callback contexts skips checking for creation of these objects
    def test_process_figure_data_frame_str_df(self, dash_table_with_str_dataframe, gapminder):
        data_manager["gapminder"] = gapminder
        table = vm.Table(id="table", figure=dash_table_with_str_dataframe)
        assert data_manager[table["data_frame"]].load().equals(gapminder)

    def test_process_figure_data_frame_df(self, standard_dash_table, gapminder):
        table = vm.Table(id="table", figure=standard_dash_table)
        assert data_manager[table["data_frame"]].load().equals(gapminder)


class TestPreBuildTable:
    def test_pre_build_no_actions_no_underlying_table_id(self, standard_dash_table):
        table = vm.Table(id="text_table", figure=standard_dash_table)
        table.pre_build()

        assert table._input_component_id == "__input_text_table"

    def test_pre_build_actions_underlying_table_id(self, dash_data_table_with_id, filter_interaction_action):
        table = vm.Table(id="text_table", figure=dash_data_table_with_id, actions=[filter_interaction_action])
        table.pre_build()

        assert table._input_component_id == "underlying_table_id"


class TestBuildTable:
    def test_table_build_mandatory_only(self, standard_dash_table, gapminder):
        table = vm.Table(id="text_table", figure=standard_dash_table)
        table.pre_build()
        table = table.build()
        expected_table = dcc.Loading(
            children=[
                None,
                html.Div(
                    id="text_table",
                    children=[html.Div(id="__input_text_table")],
                    className="table-container",
                ),
            ],
            color="grey",
            parent_className="loading-container",
            overlay_style={"visibility": "visible", "opacity": 0.3},
        )

        assert_component_equal(table, expected_table)

    def test_table_build_with_underlying_id(self, dash_data_table_with_id, filter_interaction_action, gapminder):
        table = vm.Table(id="text_table", figure=dash_data_table_with_id)
        table.pre_build()
        table = table.build()

        expected_table = dcc.Loading(
            children=[
                None,
                html.Div(
                    id="text_table",
                    children=[html.Div(id="underlying_table_id")],
                    className="table-container",
                ),
            ],
            color="grey",
            parent_className="loading-container",
            overlay_style={"visibility": "visible", "opacity": 0.3},
        )

        assert_component_equal(table, expected_table)

    def test_table_build_with_title(self, standard_dash_table, gapminder):
        table = vm.Table(id="text_table", title="Table Title", figure=standard_dash_table)
        table.pre_build()
        table = table.build()
        expected_table = dcc.Loading(
            children=[
                html.H3("Table Title"),
                html.Div(
                    id="text_table",
                    children=[html.Div(id="__input_text_table")],
                    className="table-container",
                ),
            ],
            color="grey",
            parent_className="loading-container",
            overlay_style={"visibility": "visible", "opacity": 0.3},
        )

        assert_component_equal(table, expected_table)



import pytest
import vizro.models as vm
import vizro.plotly.express as px
from dash._callback_context import context_value
from dash._utils import AttributeDict
from vizro._constants import PARAMETER_ACTION_PREFIX
from vizro.actions._actions_utils import CallbackTriggerDict
from vizro.managers import data_manager, model_manager


@pytest.fixture
def target_scatter_parameter_y(request, gapminder_2007, scatter_params):
    y = request.param
    scatter_params["y"] = y
    return px.scatter(gapminder_2007, **scatter_params).update_layout(
        margin_t=24, title_pad_l=0, title_pad_r=0, margin_l=24
    )


@pytest.fixture
def target_scatter_matrix_parameter_dimensions(request, iris, scatter_matrix_params):
    dimensions = request.param
    scatter_matrix_params["dimensions"] = dimensions
    return px.scatter_matrix(iris, **scatter_matrix_params).update_layout(
        margin_t=24, title_pad_l=0, title_pad_r=0, margin_l=24
    )


@pytest.fixture
def target_scatter_parameter_hover_data(request, gapminder_2007, scatter_params):
    hover_data = request.param
    scatter_params["hover_data"] = hover_data
    return px.scatter(gapminder_2007, **scatter_params).update_layout(
        margin_t=24, title_pad_l=0, title_pad_r=0, margin_l=24
    )


@pytest.fixture
def target_box_parameter_y(request, gapminder_2007, box_params):
    y = request.param
    box_params["y"] = y
    return px.box(gapminder_2007, **box_params).update_layout(margin_t=24, title_pad_l=0, title_pad_r=0, margin_l=24)


@pytest.fixture
def target_scatter_parameter_y_and_x(request, gapminder_2007, scatter_params):
    y, x = request.param
    scatter_params["y"] = y
    scatter_params["x"] = x
    return px.scatter(gapminder_2007, **scatter_params).update_layout(
        margin_t=24, title_pad_l=0, title_pad_r=0, margin_l=24
    )


@pytest.fixture
def target_scatter_parameter_data_frame_first_n_last_n(
    request, gapminder_dynamic_first_n_last_n_function, scatter_params
):
    first_n_last_n_args = request.param
    return px.scatter(gapminder_dynamic_first_n_last_n_function(**first_n_last_n_args), **scatter_params).update_layout(
        margin_t=24, title_pad_l=0, title_pad_r=0, margin_l=24
    )


@pytest.fixture
def target_box_parameter_data_frame_first_n_last_n(request, gapminder_dynamic_first_n_last_n_function, box_params):
    first_n_last_n_args = request.param
    return px.box(gapminder_dynamic_first_n_last_n_function(**first_n_last_n_args), **box_params).update_layout(
        margin_t=24, title_pad_l=0, title_pad_r=0, margin_l=24
    )


@pytest.fixture
def target_box_parameter_y_and_x(request, gapminder_2007, box_params):
    y, x = request.param
    box_params["y"] = y
    box_params["x"] = x
    return px.box(gapminder_2007, **box_params).update_layout(margin_t=24, title_pad_l=0, title_pad_r=0, margin_l=24)


@pytest.fixture
def ctx_parameter_y(request):
    """Mock dash.ctx that represents y-axis Parameter value selection."""
    y = request.param
    mock_ctx = {
        "args_grouping": {
            "external": {
                "filter_interaction": [],
                "filters": [],
                "parameters": [
                    CallbackTriggerDict(
                        id="y_parameter",
                        property="value",
                        value=y,
                        str_id="y_parameter",
                        triggered=False,
                    )
                ],
                "theme_selector": CallbackTriggerDict(
                    id="theme_selector",
                    property="checked",
                    value=False,
                    str_id="theme_selector",
                    triggered=False,
                ),
            }
        }
    }
    context_value.set(AttributeDict(**mock_ctx))
    return context_value


@pytest.fixture
def ctx_parameter_dimensions(request):
    """Mock dash.ctx that represents `dimensions` Parameter value selection."""
    y = request.param
    mock_ctx = {
        "args_grouping": {
            "external": {
                "filter_interaction": [],
                "filters": [],
                "parameters": [
                    CallbackTriggerDict(
                        id="dimensions_parameter",
                        property="value",
                        value=y,
                        str_id="dimensions_parameter",
                        triggered=False,
                    )
                ],
                "theme_selector": CallbackTriggerDict(
                    id="theme_selector",
                    property="checked",
                    value=False,
                    str_id="theme_selector",
                    triggered=False,
                ),
            }
        }
    }
    context_value.set(AttributeDict(**mock_ctx))
    return context_value


@pytest.fixture
def ctx_parameter_hover_data(request):
    """Mock dash.ctx that represents hover_data Parameter value selection."""
    hover_data = request.param
    mock_ctx = {
        "args_grouping": {
            "external": {
                "filter_interaction": [],
                "filters": [],
                "parameters": [
                    CallbackTriggerDict(
                        id="hover_data_parameter",
                        property="value",
                        value=hover_data,
                        str_id="hover_data_parameter",
                        triggered=False,
                    )
                ],
                "theme_selector": CallbackTriggerDict(
                    id="theme_selector",
                    property="checked",
                    value=False,
                    str_id="theme_selector",
                    triggered=False,
                ),
            }
        }
    }
    context_value.set(AttributeDict(**mock_ctx))
    return context_value


@pytest.fixture
def ctx_parameter_y_and_x(request):
    """Mock dash.ctx that represents y-axis Parameter value selection."""
    y, x = request.param
    mock_ctx = {
        "args_grouping": {
            "external": {
                "filter_interaction": [],
                "filters": [],
                "parameters": [
                    CallbackTriggerDict(
                        id="y_parameter",
                        property="value",
                        value=y,
                        str_id="y_parameter",
                        triggered=False,
                    ),
                    CallbackTriggerDict(
                        id="x_parameter",
                        property="value",
                        value=x,
                        str_id="x_parameter",
                        triggered=False,
                    ),
                ],
                "theme_selector": CallbackTriggerDict(
                    id="theme_selector",
                    property="checked",
                    value=False,
                    str_id="theme_selector",
                    triggered=False,
                ),
            }
        }
    }
    context_value.set(AttributeDict(**mock_ctx))
    return context_value


@pytest.fixture
def ctx_parameter_data_frame_argument(request):
    """Mock dash.ctx that represents parameter applied."""
    targets, first_n_last_n_args = request.param

    parameters = [
        CallbackTriggerDict(
            id="first_n_parameter",
            property="value",
            value=first_n_last_n_args["first_n"],
            str_id="first_n_parameter",
            triggered=False,
        )
    ]

    if last_n := first_n_last_n_args.get("last_n"):
        parameters.append(
            CallbackTriggerDict(
                id="last_n_parameter",
                property="value",
                value=last_n,
                str_id="last_n_parameter",
                triggered=False,
            )
        )

    mock_ctx = {
        "args_grouping": {"external": {"filters": [], "filter_interaction": [], "parameters": parameters}},
        "outputs_list": [
            {"id": {"action_id": "test_action", "target_id": target, "type": "download_dataframe"}, "property": "data"}
            for target in targets
        ],
    }

    context_value.set(AttributeDict(**mock_ctx))
    return context_value


class TestParameter:
    @pytest.mark.usefixtures("managers_one_page_two_graphs_one_button")
    @pytest.mark.parametrize(
        "ctx_parameter_y, target_scatter_parameter_y",
        [("pop", "pop"), ("gdpPercap", "gdpPercap"), ("NONE", None)],
        indirect=True,
    )
    def test_one_parameter_one_target(self, ctx_parameter_y, target_scatter_parameter_y):
        # Creating and adding a Parameter object to the existing Page
        y_parameter = vm.Parameter(
            id="test_parameter",
            targets=["scatter_chart.y"],
            selector=vm.RadioItems(id="y_parameter", options=["lifeExp", "pop", "gdpPercap"], value="lifeExp"),
        )
        model_manager["test_page"].controls = [y_parameter]

        # Adds a default _parameter Action to the "y_parameter" selector object
        y_parameter.pre_build()

        # Run action by picking the above added action function and executing it with ()
        result = model_manager[f"{PARAMETER_ACTION_PREFIX}_test_parameter"].function()
        expected = {"scatter_chart": target_scatter_parameter_y}

        assert result == expected

    @pytest.mark.usefixtures("managers_one_page_two_graphs_one_button")
    @pytest.mark.parametrize(
        "ctx_parameter_hover_data, target_scatter_parameter_hover_data",
        [
            (["NONE"], [None]),
            (["NONE", "pop"], ["pop"]),
            (["NONE", "NONE"], [None]),
            (["ALL"], ["lifeExp", "pop", "gdpPercap"]),
            (["ALL", "lifeExp"], ["lifeExp", "pop", "gdpPercap"]),
            (["ALL", "NONE"], ["lifeExp", "pop", "gdpPercap"]),
        ],
        indirect=True,
    )
    def test_one_parameter_one_target_NONE_list(self, ctx_parameter_hover_data, target_scatter_parameter_hover_data):
        # Creating and adding a Parameter object to the existing Page
        y_parameter = vm.Parameter(
            id="test_parameter",
            targets=["scatter_chart.hover_data"],
            selector=vm.Dropdown(
                id="hover_data_parameter", multi=True, options=["NONE", "lifeExp", "pop", "gdpPercap"], value=["NONE"]
            ),
        )
        model_manager["test_page"].controls = [y_parameter]

        # Adds a default _parameter Action to the "y_parameter" selector object
        y_parameter.pre_build()

        # Run action by picking the above added action function and executing it with ()
        result = model_manager[f"{PARAMETER_ACTION_PREFIX}_test_parameter"].function()
        expected = {"scatter_chart": target_scatter_parameter_hover_data}

        assert result == expected

    @pytest.mark.usefixtures("managers_one_page_two_graphs_one_button")
    @pytest.mark.parametrize(
        "ctx_parameter_y, target_scatter_parameter_y, target_box_parameter_y",
        [("pop", "pop", "pop"), ("gdpPercap", "gdpPercap", "gdpPercap")],
        indirect=True,
    )
    def test_one_parameter_multiple_targets(self, ctx_parameter_y, target_scatter_parameter_y, target_box_parameter_y):
        # Creating and adding a Parameter object to the existing Page
        y_parameter = vm.Parameter(
            id="test_parameter",
            targets=["scatter_chart.y", "box_chart.y"],
            selector=vm.RadioItems(id="y_parameter", options=["lifeExp", "pop", "gdpPercap"], value="lifeExp"),
        )
        model_manager["test_page"].controls = [y_parameter]

        # Adds a default _parameter Action to the "y_parameter" selector object
        y_parameter.pre_build()

        # Run action by picking the above added action function and executing it with ()
        result = model_manager[f"{PARAMETER_ACTION_PREFIX}_test_parameter"].function()
        expected = {"scatter_chart": target_scatter_parameter_y, "box_chart": target_box_parameter_y}

        assert result == expected

    @pytest.mark.usefixtures("managers_one_page_two_graphs_one_button")
    @pytest.mark.parametrize(
        "ctx_parameter_y_and_x, target_scatter_parameter_y_and_x",
        [(["pop", "continent"], ["pop", "continent"]), (["gdpPercap", "country"], ["gdpPercap", "country"])],
        indirect=True,
    )
    def test_multiple_parameters_one_target(self, ctx_parameter_y_and_x, target_scatter_parameter_y_and_x):
        # Creating and adding a Parameter object to the existing Page
        y_parameter = vm.Parameter(
            id="test_parameter_x",
            targets=["scatter_chart.y"],
            selector=vm.RadioItems(id="y_parameter", options=["lifeExp", "pop", "gdpPercap"], value="lifeExp"),
        )
        x_parameter = vm.Parameter(
            id="test_parameter_y",
            targets=["scatter_chart.x"],
            selector=vm.RadioItems(id="x_parameter", options=["continent", "country"], value="continent"),
        )
        model_manager["test_page"].controls = [y_parameter, x_parameter]

        # Adds a default _parameter Action to the "y_parameter" selector object
        y_parameter.pre_build()
        x_parameter.pre_build()

        # Run action by picking the above added action function and executing it with ()
        result = model_manager[f"{PARAMETER_ACTION_PREFIX}_test_parameter_x"].function()
        expected = {"scatter_chart": target_scatter_parameter_y_and_x}

        assert result == expected

    @pytest.mark.usefixtures("managers_one_page_two_graphs_one_button")
    @pytest.mark.parametrize(
        "ctx_parameter_y_and_x, target_scatter_parameter_y_and_x, target_box_parameter_y_and_x",
        [
            (["pop", "continent"], ["pop", "continent"], ["pop", "continent"]),
            (["gdpPercap", "country"], ["gdpPercap", "country"], ["gdpPercap", "country"]),
        ],
        indirect=True,
    )
    def test_multiple_parameters_multiple_targets(
        self, ctx_parameter_y_and_x, target_scatter_parameter_y_and_x, target_box_parameter_y_and_x
    ):
        # Creating and adding a Parameter object to the existing Page
        y_parameter = vm.Parameter(
            id="test_parameter_x",
            targets=["scatter_chart.y", "box_chart.y"],
            selector=vm.RadioItems(id="y_parameter", options=["lifeExp", "pop", "gdpPercap"], value="lifeExp"),
        )
        x_parameter = vm.Parameter(
            id="test_parameter_y",
            targets=["scatter_chart.x", "box_chart.x"],
            selector=vm.RadioItems(id="x_parameter", options=["continent", "country"], value="continent"),
        )
        model_manager["test_page"].controls = [y_parameter, x_parameter]

        # Adds a default _parameter Action to the "y_parameter" selector object
        y_parameter.pre_build()
        x_parameter.pre_build()

        # Run action by picking the above added action function and executing it with ()
        result = model_manager[f"{PARAMETER_ACTION_PREFIX}_test_parameter_x"].function()
        expected = {"scatter_chart": target_scatter_parameter_y_and_x, "box_chart": target_box_parameter_y_and_x}

        assert result == expected

    @pytest.mark.usefixtures("managers_one_page_two_graphs_one_button")
    @pytest.mark.parametrize(
        "ctx_parameter_y_and_x, target_scatter_parameter_y_and_x, target_box_parameter_y_and_x",
        [
            (["pop", "continent"], ["pop", "pop"], ["continent", "continent"]),
            (["gdpPercap", "country"], ["gdpPercap", "gdpPercap"], ["country", "country"]),
        ],
        indirect=True,
    )
    def test_one_parameter_per_target_multiple_attributes(
        self, ctx_parameter_y_and_x, target_scatter_parameter_y_and_x, target_box_parameter_y_and_x
    ):
        # Creating and adding a Parameter object to the existing Page
        scatter_parameter = vm.Parameter(
            id="test_parameter_scatter",
            targets=["scatter_chart.y", "scatter_chart.x"],
            selector=vm.RadioItems(id="y_parameter", options=["lifeExp", "pop", "gdpPercap"], value="lifeExp"),
        )
        box_parameter = vm.Parameter(
            id="test_parameter_box",
            targets=["box_chart.y", "box_chart.x"],
            selector=vm.RadioItems(id="x_parameter", options=["continent", "country"], value="continent"),
        )
        model_manager["test_page"].controls = [scatter_parameter, box_parameter]

        # Adds a default _parameter Action to the "y_parameter" selector object
        scatter_parameter.pre_build()
        box_parameter.pre_build()

        # Run action by picking the above added action function and executing it with ()
        result = model_manager[f"{PARAMETER_ACTION_PREFIX}_test_parameter_scatter"].function()
        expected = {"scatter_chart": target_scatter_parameter_y_and_x}

        assert result == expected

        result = model_manager[f"{PARAMETER_ACTION_PREFIX}_test_parameter_box"].function()
        expected = {"box_chart": target_box_parameter_y_and_x}

        assert result == expected

    @pytest.mark.usefixtures("managers_one_page_two_graphs_with_dynamic_data")
    @pytest.mark.parametrize(
        "ctx_parameter_data_frame_argument, " "target_scatter_parameter_data_frame_first_n_last_n",
        [
            ((["scatter_chart"], {"first_n": 50}), {"first_n": 50}),
            ((["scatter_chart"], {"first_n": 50, "last_n": 50}), {"first_n": 50, "last_n": 50}),
        ],
        indirect=[
            "ctx_parameter_data_frame_argument",
            "target_scatter_parameter_data_frame_first_n_last_n",
        ],
    )
    def test_data_frame_parameters_one_target(
        self,
        ctx_parameter_data_frame_argument,
        target_scatter_parameter_data_frame_first_n_last_n,
        gapminder_dynamic_first_n_last_n_function,
    ):
        data_manager["gapminder_dynamic_first_n_last_n"] = gapminder_dynamic_first_n_last_n_function

        # Creating and adding a Parameter object (data_frame function argument parametrizing) to the existing Page
        first_n_parameter = vm.Parameter(
            id="test_data_frame_parameter",
            targets=["scatter_chart.data_frame.first_n"],
            selector=vm.Slider(id="first_n_parameter", min=1, max=10, step=1),
        )
        model_manager["test_page"].controls.append(first_n_parameter)
        first_n_parameter.pre_build()

        # This parameter will affect results only if included in the ctx_parameter_data_frame_argument:
        last_n_parameter = vm.Parameter(
            id="test_data_frame_parameter_last_n",
            targets=["scatter_chart.data_frame.last_n"],
            selector=vm.Slider(id="last_n_parameter", min=1, max=10, step=1),
        )
        model_manager["test_page"].controls.append(last_n_parameter)
        last_n_parameter.pre_build()

        # Run action by picking the above added action function and executing it with ()
        result = model_manager[f"{PARAMETER_ACTION_PREFIX}_test_data_frame_parameter"].function()

        expected = {
            "scatter_chart": target_scatter_parameter_data_frame_first_n_last_n,
        }

        assert result == expected

    @pytest.mark.usefixtures("managers_one_page_two_graphs_with_dynamic_data")
    @pytest.mark.parametrize(
        "ctx_parameter_data_frame_argument, "
        "target_scatter_parameter_data_frame_first_n_last_n, "
        "target_box_parameter_data_frame_first_n_last_n",
        [
            (
                (["scatter_chart", "box_chart"], {"first_n": 50}),
                {"first_n": 50},
                {"first_n": 50},
            ),
            (
                (["scatter_chart", "box_chart"], {"first_n": 50, "last_n": 50}),
                {"first_n": 50, "last_n": 50},
                {"first_n": 50, "last_n": 50},
            ),
        ],
        indirect=[
            "ctx_parameter_data_frame_argument",
            "target_scatter_parameter_data_frame_first_n_last_n",
            "target_box_parameter_data_frame_first_n_last_n",
        ],
    )
    def test_data_frame_parameters_multiple_targets(
        self,
        ctx_parameter_data_frame_argument,
        target_scatter_parameter_data_frame_first_n_last_n,
        target_box_parameter_data_frame_first_n_last_n,
        gapminder_dynamic_first_n_last_n_function,
    ):
        data_manager["gapminder_dynamic_first_n_last_n"] = gapminder_dynamic_first_n_last_n_function

        # Creating and adding a Parameter object (data_frame function argument parametrizing) to the existing Page
        first_n_parameter = vm.Parameter(
            id="test_data_frame_parameter",
            targets=["scatter_chart.data_frame.first_n", "box_chart.data_frame.first_n"],
            selector=vm.Slider(id="first_n_parameter", min=1, max=10, step=1),
        )
        model_manager["test_page"].controls.append(first_n_parameter)
        first_n_parameter.pre_build()

        # This parameter will affect results only if included in the ctx_parameter_data_frame_argument:
        last_n_parameter = vm.Parameter(
            id="test_data_frame_parameter_last_n",
            targets=["scatter_chart.data_frame.last_n", "box_chart.data_frame.last_n"],
            selector=vm.Slider(id="last_n_parameter", min=1, max=10, step=1),
        )
        model_manager["test_page"].controls.append(last_n_parameter)
        last_n_parameter.pre_build()

        # Run action by picking the above added action function and executing it with ()
        result = model_manager[f"{PARAMETER_ACTION_PREFIX}_test_data_frame_parameter"].function()

        expected = {
            "scatter_chart": target_scatter_parameter_data_frame_first_n_last_n,
            "box_chart": target_box_parameter_data_frame_first_n_last_n,
        }

        assert result == expected

    @pytest.mark.usefixtures("managers_one_page_one_graph_with_dict_param_input")
    @pytest.mark.parametrize(
        "ctx_parameter_dimensions, target_scatter_matrix_parameter_dimensions",
        [("ALL", ["sepal_length", "sepal_width", "petal_length", "petal_width"]), (["sepal_width"], ["sepal_width"])],
        indirect=True,
    )
    def test_one_parameter_with_dict_input_as_options(
        self, ctx_parameter_dimensions, target_scatter_matrix_parameter_dimensions
    ):
        # If the options are provided as a list of dictionaries, the value should be correctly passed to the
        # target as a list. So when "ALL" is selected, a list of all possible values should be returned.
        dimensions_parameter = vm.Parameter(
            id="test_parameter_dimensions",
            targets=["scatter_matrix_chart.dimensions"],
            selector=vm.RadioItems(
                id="dimensions_parameter",
                options=[
                    {"label": "sepal_length", "value": "sepal_length"},
                    {"label": "sepal_width", "value": "sepal_width"},
                    {"label": "petal_length", "value": "petal_length"},
                    {"label": "petal_width", "value": "petal_width"},
                ],
            ),
        )
        model_manager["test_page"].controls = [dimensions_parameter]
        dimensions_parameter.pre_build()

        # Run action by picking the above added action function and executing it with ()
        result = model_manager[f"{PARAMETER_ACTION_PREFIX}_test_parameter_dimensions"].function()
        expected = {"scatter_matrix_chart": target_scatter_matrix_parameter_dimensions}

        assert result == expected



import vizro.models as vm
import vizro.plotly.express as px
from vizro import Vizro

stocks = px.data.stocks()

page = vm.Page(
    title="Line",
    components=[vm.Graph(figure=px.line(stocks, x="date", y="GOOG"))],
)

dashboard = vm.Dashboard(pages=[page])
Vizro().build(dashboard).run()



import vizro.models as vm
import vizro.plotly.express as px
from vizro import Vizro

tips = px.data.tips()


page = vm.Page(
    title="Stacked bar",
    components=[vm.Graph(figure=px.histogram(tips, y="sex", x="total_bill", color="day", orientation="h"))],
)

dashboard = vm.Dashboard(pages=[page])
Vizro().build(dashboard).run()



"""Unit tests for vizro.managers.data_manager."""

from contextlib import suppress
from functools import partial

import numpy as np
import pandas as pd
import pytest
from asserts import assert_frame_not_equal
from flask_caching import Cache
from freezegun import freeze_time
from pandas.testing import assert_frame_equal
from vizro import Vizro
from vizro.managers import data_manager


# Fixture that freezes the time so that tests involving time.sleep can run quickly. Instead of time.sleep,
# tests should use freezer.tick() to advance time. This is very similar to the fixture in the pytest-freezegun package.
# This makes it possible to test with realistic timeouts that flask-caching can handle well. Otherwise we
# test with very low timeouts and time.sleep(1), and this is flaky since flask-caching is not designed to
# handle such small intervals (e.g. it rounds times to the nearest second).
# We use tick=True so that time continues to pass between directly consecutive calls to a load function. This makes the
# behavior in tests as close as possible to real world.
@pytest.fixture
def freezer():
    with freeze_time(tick=True) as frozen_time:
        yield frozen_time


@pytest.fixture(autouse=True)
def clear_cache():
    yield
    # Vizro._reset doesn't empty the cache, so any tests which have something other than NullCache must clear it
    # after running. Suppress AttributeError: 'Cache' object has no attribute 'app' that occurs when
    # data_manager._cache_has_app is False.
    with suppress(AttributeError):
        data_manager.cache.clear()


def make_fixed_data():
    return pd.DataFrame([1, 2, 3])


class TestLoad:
    def test_static(self):
        data = make_fixed_data()
        data_manager["data"] = data
        loaded_data = data_manager["data"].load()
        assert_frame_equal(loaded_data, data)
        # Make sure loaded_data is a copy rather than the same object.
        assert loaded_data is not data

    def test_dynamic(self):
        data = make_fixed_data
        data_manager["data"] = data
        loaded_data = data_manager["data"].load()
        assert_frame_equal(loaded_data, data())
        # Make sure loaded_data is a copy rather than the same object.
        assert loaded_data is not data()

    def test_dynamic_lambda(self):
        data = lambda: make_fixed_data()  # noqa: E731
        data_manager["data"] = data
        loaded_data = data_manager["data"].load()
        assert_frame_equal(loaded_data, data())
        # Make sure loaded_data is a copy rather than the same object.
        assert loaded_data is not data()


class TestInvalid:
    def test_static_data_does_not_support_timeout(self):
        data = make_fixed_data()
        data_manager["data"] = data
        with pytest.raises(
            AttributeError, match="Static data that is a pandas.DataFrame itself does not support timeout"
        ):
            data_manager["data"].timeout = 10

    def test_setitem_invalid_type(self):
        with pytest.raises(
            TypeError, match="Data source data must be a pandas DataFrame or function that returns a pandas DataFrame."
        ):
            data_manager["data"] = pd.Series([1, 2, 3])

    def test_does_not_exist(self):
        with pytest.raises(KeyError, match="Data source data does not exist."):
            data_manager["data"]


def make_random_data():
    return pd.DataFrame(np.random.default_rng().random(3))


def make_random_data_with_args(label="x"):
    return make_random_data().assign(label=label)


# This is important to test since it's like how kedro datasets work.
class RandomData:
    # This cannot be @staticmethod since we want to test it as a bound method.
    def load(self):
        return make_random_data()


class RandomDataWithArgs:
    # This cannot be @staticmethod since we want to test it as a bound method.
    def load(self, label="x"):
        return make_random_data_with_args(label)


make_random_data_lambda = lambda: make_random_data()  # noqa: E731

make_random_data_partial = partial(make_random_data)

make_random_data_with_args_lambda = lambda label="x": make_random_data_with_args(label)  # noqa: E731

make_random_data_with_args_partial = partial(make_random_data_with_args)


# We test the function and bound method cases but not the unbound methods RandomData.load and RandomDataWithArgs.load
# which are not important.
@pytest.mark.parametrize(
    "data_callable",
    [
        make_random_data,
        make_random_data_with_args,
        RandomData().load,
        RandomDataWithArgs().load,
        make_random_data_lambda,
        make_random_data_with_args_lambda,
        make_random_data_partial,
        make_random_data_with_args_partial,
    ],
)
class TestCacheNotOperational:
    def test_null_cache_no_app(self, data_callable):
        # No app at all, so data_manager._cache_has_app is False.
        data_manager["data"] = data_callable
        loaded_data_1 = data_manager["data"].load()
        loaded_data_2 = data_manager["data"].load()
        assert_frame_not_equal(loaded_data_1, loaded_data_2)

    def test_null_cache_with_app(self, data_callable):
        # App exists but cache is NullCache so does not do anything.
        data_manager["data"] = data_callable
        Vizro()
        loaded_data_1 = data_manager["data"].load()
        loaded_data_2 = data_manager["data"].load()
        assert_frame_not_equal(loaded_data_1, loaded_data_2)

    def test_cache_no_app(self, data_callable):
        # App exists and has a real cache but data_manager.cache is set too late so app is not attached to cache.
        data_manager["data"] = data_callable
        Vizro()
        data_manager.cache = Cache(config={"CACHE_TYPE": "SimpleCache"})

        with pytest.warns(UserWarning, match="Cache does not have Vizro app attached and so is not operational."):
            loaded_data_1 = data_manager["data"].load()
            loaded_data_2 = data_manager["data"].load()
        assert_frame_not_equal(loaded_data_1, loaded_data_2)


@pytest.fixture
def simple_cache():
    # We don't need the Flask request context to run tests. (flask-caching tests for memoize use
    # app.test_request_context() but look like they don't actually need to, since only flask_caching.Cache.cached
    # requires the request context.)
    # We do need a Flask app to be attached for the cache to be operational though, hence the call Vizro().
    data_manager.cache = Cache(config={"CACHE_TYPE": "SimpleCache"})
    Vizro()
    yield


# We test the function and bound method cases but not the unbound method RandomData.load which is not important.
@pytest.mark.parametrize(
    "data_callable",
    [
        make_random_data,
        RandomData().load,
        make_random_data_lambda,
        make_random_data_partial,
    ],
)
class TestCache:
    def test_default_timeout(self, data_callable, simple_cache, freezer):
        data_manager["data"] = data_callable

        loaded_data_1 = data_manager["data"].load()
        loaded_data_2 = data_manager["data"].load()
        # Default timeout is 300, so wait for longer than that.
        freezer.tick(300 + 50)
        loaded_data_3 = data_manager["data"].load()
        loaded_data_4 = data_manager["data"].load()

        # Cache has expired between loaded_data_2 and loaded_data_3 only.
        assert_frame_equal(loaded_data_1, loaded_data_2)
        assert_frame_equal(loaded_data_3, loaded_data_4)
        assert_frame_not_equal(loaded_data_2, loaded_data_3)

    def test_change_non_default_timeout(self, data_callable, freezer):
        data_manager.cache = Cache(config={"CACHE_TYPE": "SimpleCache", "CACHE_DEFAULT_TIMEOUT": 100})
        Vizro()
        data_manager["data"] = data_callable

        loaded_data_1 = data_manager["data"].load()
        loaded_data_2 = data_manager["data"].load()
        freezer.tick(100 + 50)
        loaded_data_3 = data_manager["data"].load()
        loaded_data_4 = data_manager["data"].load()

        # Cache has expired between loaded_data_2 and loaded_data_3 only.
        assert_frame_equal(loaded_data_1, loaded_data_2)
        assert_frame_equal(loaded_data_3, loaded_data_4)
        assert_frame_not_equal(loaded_data_2, loaded_data_3)

    def test_change_individual_timeout(self, data_callable, simple_cache, freezer):
        data_manager["data"] = data_callable
        data_manager["data"].timeout = 100

        loaded_data_1 = data_manager["data"].load()
        loaded_data_2 = data_manager["data"].load()
        freezer.tick(100 + 50)
        loaded_data_3 = data_manager["data"].load()
        loaded_data_4 = data_manager["data"].load()

        # Cache has expired between loaded_data_2 and loaded_data_3 only.
        assert_frame_equal(loaded_data_1, loaded_data_2)
        assert_frame_equal(loaded_data_3, loaded_data_4)
        assert_frame_not_equal(loaded_data_2, loaded_data_3)


# We test the function and bound method cases but not the unbound method RandomDataWithArgs.load which is not important.
@pytest.mark.usefixtures("simple_cache")
@pytest.mark.parametrize(
    "data_callable",
    [
        make_random_data_with_args,
        RandomDataWithArgs().load,
        make_random_data_with_args_lambda,
        make_random_data_with_args_partial,
    ],
)
class TestCacheWithArguments:
    def test_default_timeout(self, data_callable, freezer):
        # Analogous to TestCache.test_default_timeout
        data_manager["data"] = data_callable

        loaded_data_x_1 = data_manager["data"].load("x")
        loaded_data_x_2 = data_manager["data"].load("x")
        loaded_data_y_1 = data_manager["data"].load("y")
        loaded_data_y_2 = data_manager["data"].load("y")
        # Default timeout is 300, so wait for longer than that.
        freezer.tick(300 + 50)
        loaded_data_x_3 = data_manager["data"].load("x")
        loaded_data_x_4 = data_manager["data"].load("x")
        loaded_data_y_3 = data_manager["data"].load("y")
        loaded_data_y_4 = data_manager["data"].load("y")

        # For both x and y, cache has expired between loaded_data_2 and loaded_data_3 only.
        assert_frame_equal(loaded_data_x_1, loaded_data_x_2)
        assert_frame_equal(loaded_data_x_3, loaded_data_x_4)
        assert_frame_not_equal(loaded_data_x_2, loaded_data_x_3)

        assert_frame_equal(loaded_data_y_1, loaded_data_y_2)
        assert_frame_equal(loaded_data_y_3, loaded_data_y_4)
        assert_frame_not_equal(loaded_data_y_2, loaded_data_y_3)

        assert_frame_not_equal(loaded_data_x_1, loaded_data_y_1)
        assert_frame_not_equal(loaded_data_x_3, loaded_data_y_3)

    def test_change_individual_timeout(self, data_callable, freezer):
        # Analogous to TestCache.test_change_individual_timeout.
        data_manager["data"] = data_callable
        data_manager["data"].timeout = 100

        loaded_data_x_1 = data_manager["data"].load("x")
        loaded_data_x_2 = data_manager["data"].load("x")
        loaded_data_y_1 = data_manager["data"].load("y")
        loaded_data_y_2 = data_manager["data"].load("y")
        freezer.tick(100 + 50)
        loaded_data_x_3 = data_manager["data"].load("x")
        loaded_data_x_4 = data_manager["data"].load("x")
        loaded_data_y_3 = data_manager["data"].load("y")
        loaded_data_y_4 = data_manager["data"].load("y")

        # For both x and y, cache has expired between loaded_data_2 and loaded_data_3 only.
        assert_frame_equal(loaded_data_x_1, loaded_data_x_2)
        assert_frame_equal(loaded_data_x_3, loaded_data_x_4)
        assert_frame_not_equal(loaded_data_x_2, loaded_data_x_3)

        assert_frame_equal(loaded_data_y_1, loaded_data_y_2)
        assert_frame_equal(loaded_data_y_3, loaded_data_y_4)
        assert_frame_not_equal(loaded_data_y_2, loaded_data_y_3)

        assert_frame_not_equal(loaded_data_x_1, loaded_data_y_1)
        assert_frame_not_equal(loaded_data_x_3, loaded_data_y_3)

    def test_timeout_expires_all(self, data_callable, freezer):
        # When the cache for one set of memoized arguments expires, the cache for the whole data source expires, even
        # for other values of memoized arguments.
        # This behavior is not particularly desirable (in fact it's maybe a bit annoying); the test is here just
        # to document the current behavior. It's not easy to change this behavior within flask-caching.
        # Remember the default timeout is 300s.
        # Loading sequence of data sources is as follows:
        # t=0:      load x_1
        # t=200:    load x_2     y_1  -> x cache has not expired
        # t=400:    load x_3     y_2  -> x cache has expired. y cache might be expected to not expire but also has.
        # t=600:    load         y_3
        data_manager["data"] = data_callable

        loaded_data_x_1 = data_manager["data"].load("x")
        freezer.tick(150 + 50)
        loaded_data_x_2 = data_manager["data"].load("x")
        loaded_data_y_1 = data_manager["data"].load("y")
        freezer.tick(150 + 50)
        loaded_data_x_3 = data_manager["data"].load("x")
        loaded_data_y_2 = data_manager["data"].load("y")
        freezer.tick(150 + 50)
        loaded_data_y_3 = data_manager["data"].load("y")

        # These are as you would expect.
        assert_frame_equal(loaded_data_x_1, loaded_data_x_2)
        assert_frame_not_equal(loaded_data_x_2, loaded_data_x_3)

        # These you might expect to be the other way round:
        # assert_frame_equal(loaded_data_y_1, loaded_data_y_2)
        # assert_frame_not_equal(loaded_data_y_2, loaded_data_y_3)
        assert_frame_not_equal(loaded_data_y_1, loaded_data_y_2)
        assert_frame_equal(loaded_data_y_2, loaded_data_y_3)

    def test_named_and_default_args(self, data_callable):
        data_manager["data"] = data_callable

        loaded_data_x_1 = data_manager["data"].load(label="x")
        loaded_data_x_2 = data_manager["data"].load()
        loaded_data_x_3 = data_manager["data"].load("x")

        assert_frame_equal(loaded_data_x_1, loaded_data_x_2)
        assert_frame_equal(loaded_data_x_2, loaded_data_x_3)


class TestCacheIndependence:
    # Test both data callable with and without args in one test. The ones which don't take args are just passed an
    # empty dictionary so it's like doing data_manager["data_x"].load() with no arguments.
    @pytest.mark.parametrize(
        "data_callable, kwargs",
        [
            (make_random_data, {}),
            (RandomData().load, {}),
            (make_random_data_lambda, {}),
            (make_random_data_partial, {}),
            (make_random_data_with_args, {"label": "z"}),
            (RandomDataWithArgs().load, {"label": "z"}),
            (make_random_data_with_args_lambda, {"label": "z"}),
            (make_random_data_with_args_partial, {"label": "z"}),
        ],
    )
    def test_shared_dynamic_data_callable_no_timeout(self, data_callable, kwargs, simple_cache):
        # Two data sources that share the same function or bound method are independent when neither times out.
        # It doesn't really matter if this test passes; it's mainly here just to document the current behavior. The use
        # cases for actually wanting to do this seem limited.
        data_manager["data_x"] = data_callable
        data_manager["data_y"] = data_callable

        loaded_data_x_1 = data_manager["data_x"].load(**kwargs)
        loaded_data_y_1 = data_manager["data_y"].load(**kwargs)
        loaded_data_x_2 = data_manager["data_x"].load(**kwargs)
        loaded_data_y_2 = data_manager["data_y"].load(**kwargs)

        assert_frame_equal(loaded_data_x_1, loaded_data_x_2)
        assert_frame_equal(loaded_data_y_1, loaded_data_y_2)
        assert_frame_not_equal(loaded_data_x_1, loaded_data_y_1)

    # Test both data callable with and without args in one test. The ones which don't take args are just passed an
    # empty dictionary so it's like doing data_manager["data_x"].load() with no arguments.
    @pytest.mark.parametrize(
        "data_callable, kwargs",
        [
            (make_random_data, {}),
            (RandomData().load, {}),
            (make_random_data_lambda, {}),
            (make_random_data_partial, {}),
            (make_random_data_with_args, {"label": "z"}),
            (RandomDataWithArgs().load, {"label": "z"}),
            (make_random_data_with_args_lambda, {"label": "z"}),
            (make_random_data_with_args_partial, {"label": "z"}),
        ],
    )
    def test_shared_dynamic_data_callable_with_timeout(self, data_callable, kwargs, simple_cache, freezer):
        # Two data sources that share the same function or bound method are independent when only one times out.
        # It doesn't really matter if this test passes; it's mainly here just to document the current behavior. The use
        # cases for actually wanting to do this seem limited.
        data_manager["data_x"] = data_callable
        data_manager["data_y"] = data_callable
        data_manager["data_y"].timeout = 100

        loaded_data_x_1 = data_manager["data_x"].load(**kwargs)
        loaded_data_y_1 = data_manager["data_y"].load(**kwargs)
        freezer.tick(100 + 50)
        loaded_data_x_2 = data_manager["data_x"].load(**kwargs)
        loaded_data_y_2 = data_manager["data_y"].load(**kwargs)

        # Cache has expired for data_y but not data_x.
        assert_frame_equal(loaded_data_x_1, loaded_data_x_2)
        assert_frame_not_equal(loaded_data_y_1, loaded_data_y_2)

    @pytest.mark.parametrize(
        "random_data_cls, kwargs",
        [
            (RandomData, {}),
            (RandomDataWithArgs, {"label": "z"}),
        ],
    )
    def test_independent_dynamic_data_callable_no_timeout(self, simple_cache, random_data_cls, kwargs):
        # Two data sources use same method but have *different* bound instances are independent when neither times out.
        # This is the same as test_shared_dynamic_data_callable_no_timeout but it *does* matter that this test passes.
        data_manager["data_x"] = random_data_cls().load
        data_manager["data_y"] = random_data_cls().load

        loaded_data_x_1 = data_manager["data_x"].load(**kwargs)
        loaded_data_y_1 = data_manager["data_y"].load(**kwargs)
        loaded_data_x_2 = data_manager["data_x"].load(**kwargs)
        loaded_data_y_2 = data_manager["data_y"].load(**kwargs)

        assert_frame_equal(loaded_data_x_1, loaded_data_x_2)
        assert_frame_equal(loaded_data_y_1, loaded_data_y_2)
        assert_frame_not_equal(loaded_data_x_1, loaded_data_y_1)

    @pytest.mark.parametrize(
        "random_data_cls, kwargs",
        [
            (RandomData, {}),
            (RandomDataWithArgs, {"label": "z"}),
        ],
    )
    def test_independent_dynamic_data_callable_with_timeout(self, simple_cache, random_data_cls, kwargs, freezer):
        # Two data sources use same method but have *different* bound instances are independent when only one times out.
        # This is the same as test_shared_dynamic_data_callable_with_timeout but it *does* matter that this test passes.
        data_manager["data_x"] = random_data_cls().load
        data_manager["data_y"] = random_data_cls().load
        data_manager["data_y"].timeout = 100

        loaded_data_x_1 = data_manager["data_x"].load(**kwargs)
        loaded_data_y_1 = data_manager["data_y"].load(**kwargs)
        freezer.tick(100 + 50)
        loaded_data_x_2 = data_manager["data_x"].load(**kwargs)
        loaded_data_y_2 = data_manager["data_y"].load(**kwargs)

        # Cache has expired for data_y but not data_x.
        assert_frame_equal(loaded_data_x_1, loaded_data_x_2)
        assert_frame_not_equal(loaded_data_y_1, loaded_data_y_2)



from typing import List, Literal, Optional, Union

import pytest

try:
    from pydantic.v1 import Field, ValidationError, root_validator, validator
except ImportError:  # pragma: no cov
    from pydantic import Field, ValidationError, root_validator, validator
import logging
import textwrap

import vizro.models as vm
import vizro.plotly.express as px
from typing_extensions import Annotated
from vizro.actions import export_data
from vizro.models._base import _patch_vizro_base_model_dict
from vizro.models.types import capture
from vizro.tables import dash_ag_grid


class ChildX(vm.VizroBaseModel):
    type: Literal["child_x"] = "child_x"


class ChildY(vm.VizroBaseModel):
    type: Literal["child_y"] = "child_y"


class ChildZ(vm.VizroBaseModel):
    type: Literal["child_Z"] = "child_Z"


class ChildWithForwardRef(vm.VizroBaseModel):
    type: Literal["child_with_forward_ref"] = "child_with_forward_ref"
    grandchild: "ChildXForwardRef" = None  # noqa: F821


# ChildType does not include ChildZ initially.
ChildType = Annotated[Union[ChildX, ChildY], Field(discriminator="type")]


# These parent classes must be done as fixtures so that each test gets a fresh, unmodified copy of the class.
@pytest.fixture()
def Parent():
    # e.g. Parameter.selector: SelectorType
    class _Parent(vm.VizroBaseModel):
        child: ChildType

    return _Parent


@pytest.fixture()
def ParentWithOptional():
    # e.g. Filter.selector: Optional[SelectorType]
    class _ParentWithOptional(vm.VizroBaseModel):
        child: Optional[ChildType]

    return _ParentWithOptional


@pytest.fixture()
def ParentWithList():
    # e.g. Page.controls: List[ControlType] and Page.components: List[ComponentType]
    class _ParentWithList(vm.VizroBaseModel):
        child: List[ChildType]

    return _ParentWithList


@pytest.fixture()
def ParentWithForwardRef():
    class _ParentWithForwardRef(vm.VizroBaseModel):
        child: Annotated[Union["ChildXForwardRef", "ChildYForwardRef"], Field(discriminator="type")]  # noqa: F821

    _ParentWithForwardRef.update_forward_refs(ChildXForwardRef=ChildX, ChildYForwardRef=ChildY)
    return _ParentWithForwardRef


@pytest.fixture()
def ParentWithNonDiscriminatedUnion():
    class _ParentWithNonDiscriminatedUnion(vm.VizroBaseModel):
        child: Union[ChildX, ChildY]

    return _ParentWithNonDiscriminatedUnion


class TestDiscriminatedUnion:
    def test_no_type_match(self, Parent):
        child = ChildZ()
        with pytest.raises(ValidationError, match="No match for discriminator 'type' and value 'child_Z'"):
            Parent(child=child)

    def test_add_type_model_instantiation(self, Parent):
        Parent.add_type("child", ChildZ)
        parent = Parent(child=ChildZ())
        assert isinstance(parent.child, ChildZ)

    def test_add_type_dict_instantiation(self, Parent):
        Parent.add_type("child", ChildZ)
        parent = Parent(child={"type": "child_Z"})
        assert isinstance(parent.child, ChildZ)


class TestOptionalDiscriminatedUnion:
    # Optional[ChildType] does not work correctly as a discriminated union - pydantic turns it into a regular union.
    # Hence the validation error messages are not as expected. The tests of add_type pass because in practice a
    # discriminated union is not actually needed to achieve the desired behavior. The union is still a regular one
    # even after add_type.
    @pytest.mark.xfail
    def test_no_type_match(self, ParentWithOptional):
        child = ChildZ()
        with pytest.raises(ValidationError, match="No match for discriminator 'type' and value 'child_Z'"):
            ParentWithOptional(child=child)

    # The current error message is the non-discriminated union one.
    def test_no_type_match_current_behaviour(self, ParentWithOptional):
        child = ChildZ()
        with pytest.raises(ValidationError, match="unexpected value; permitted: 'child_x'"):
            ParentWithOptional(child=child)

    def test_add_type_model_instantiation(self, ParentWithOptional):
        ParentWithOptional.add_type("child", ChildZ)
        parent = ParentWithOptional(child=ChildZ())
        assert isinstance(parent.child, ChildZ)

    def test_add_type_dict_instantiation(self, ParentWithOptional):
        ParentWithOptional.add_type("child", ChildZ)
        parent = ParentWithOptional(child={"type": "child_Z"})
        assert isinstance(parent.child, ChildZ)


class TestListDiscriminatedUnion:
    def test_no_type_match(self, ParentWithList):
        child = ChildZ()
        with pytest.raises(ValidationError, match="No match for discriminator 'type' and value 'child_Z'"):
            ParentWithList(child=[child])

    def test_add_type_model_instantiation(self, ParentWithList):
        ParentWithList.add_type("child", ChildZ)
        parent = ParentWithList(child=[ChildZ()])
        assert isinstance(parent.child[0], ChildZ)

    def test_add_type_dict_instantiation(self, ParentWithList):
        ParentWithList.add_type("child", ChildZ)
        parent = ParentWithList(child=[{"type": "child_Z"}])
        assert isinstance(parent.child[0], ChildZ)


class TestParentForwardRefDiscriminatedUnion:
    def test_no_type_match(self, ParentWithForwardRef):
        child = ChildZ()
        with pytest.raises(ValidationError, match="No match for discriminator 'type' and value 'child_Z'"):
            ParentWithForwardRef(child=child)

    def test_add_type_model_instantiation(self, ParentWithForwardRef, mocker):
        # Make it as if these are in vizro.models so that update_forward_refs call in add_type works on them.
        mocker.patch.dict(vm.__dict__, {"ChildXForwardRef": ChildX, "ChildYForwardRef": ChildY})
        ParentWithForwardRef.add_type("child", ChildZ)
        parent = ParentWithForwardRef(child=ChildZ())
        assert isinstance(parent.child, ChildZ)

    def test_add_type_dict_instantiation(self, ParentWithForwardRef, mocker):
        # Make it as if these are in vizro.models so that update_forward_refs call in add_type works on them.
        mocker.patch.dict(vm.__dict__, {"ChildXForwardRef": ChildX, "ChildYForwardRef": ChildY})
        ParentWithForwardRef.add_type("child", ChildZ)
        parent = ParentWithForwardRef(child={"type": "child_Z"})
        assert isinstance(parent.child, ChildZ)


class TestChildWithForwardRef:
    def test_no_type_match(self, Parent):
        child = ChildWithForwardRef()
        with pytest.raises(
            ValidationError, match="No match for discriminator 'type' and value 'child_with_forward_ref'"
        ):
            Parent(child=child)

    def test_add_type_model_instantiation(self, Parent, mocker):
        # Make it as if these are in vizro.models so that update_forward_refs call in add_type works on them.
        mocker.patch.dict(vm.__dict__, {"ChildXForwardRef": ChildX})
        Parent.add_type("child", ChildWithForwardRef)
        parent = Parent(child=ChildWithForwardRef(grandchild=ChildX()))
        assert isinstance(parent.child, ChildWithForwardRef) and isinstance(parent.child.grandchild, ChildX)

    def test_add_type_dict_instantiation(self, Parent, mocker):
        # Make it as if these are in vizro.models so that update_forward_refs call in add_type works on them.
        mocker.patch.dict(vm.__dict__, {"ChildXForwardRef": ChildX})
        Parent.add_type("child", ChildWithForwardRef)
        parent = Parent(child={"type": "child_with_forward_ref", "grandchild": {}})
        assert isinstance(parent.child, ChildWithForwardRef) and isinstance(parent.child.grandchild, ChildX)


def test_no_type_match(ParentWithNonDiscriminatedUnion):
    with pytest.raises(ValueError, match="Field 'child' must be a discriminated union"):
        ParentWithNonDiscriminatedUnion.add_type("child", ChildZ)


class Model(vm.VizroBaseModel):
    type: Literal["model"] = "model"


class ModelWithFieldSetting(vm.VizroBaseModel):
    type: Literal["exclude_model"] = "exclude_model"
    title: str = Field(..., description="Title to be displayed.")
    foo: str = ""

    # Set a field with regular validator
    @validator("foo", always=True)
    def set_foo(cls, foo) -> str:
        return foo or "long-random-thing"

    # Set a field with a pre=True root-validator -->
    # # this will not be caught by exclude_unset=True
    @root_validator(pre=True)
    def set_id(cls, values):
        if "title" not in values:
            return values

        values.setdefault("id", values["title"])
        return values

    # Exclude field even if missed by exclude_unset=True
    def __vizro_exclude_fields__(self):
        """Exclude id field if it is the same as the title."""
        return {"id"}


class TestDict:
    def test_dict_no_args(self):
        model = Model(id="model_id")
        assert model.dict() == {"id": "model_id", "type": "model"}

    def test_dict_exclude_unset(self):
        model = Model(id="model_id")
        assert model.dict(exclude_unset=True) == {"id": "model_id"}

    def test_dict_exclude_id(self):
        model = Model()
        assert model.dict(exclude={"id"}) == {"type": "model"}

    def test_dict_exclude_type(self):
        # __vizro_exclude_fields__ should have no effect here.
        model = Model(id="model_id")
        assert model.dict(exclude={"type"}) == {"id": "model_id"}

    def test_dict_exclude_in_model_unset_with_and_without_context(self):
        model = ModelWithFieldSetting(title="foo")
        with _patch_vizro_base_model_dict():
            assert model.dict(exclude_unset=True) == {"title": "foo", "__vizro_model__": "ModelWithFieldSetting"}
        assert model.dict(exclude_unset=True) == {"id": "foo", "title": "foo"}

    def test_dict_exclude_in_model_no_args_with_and_without_context(self):
        model = ModelWithFieldSetting(title="foo")
        with _patch_vizro_base_model_dict():
            assert model.dict() == {
                "title": "foo",
                "type": "exclude_model",
                "__vizro_model__": "ModelWithFieldSetting",
                "foo": "long-random-thing",
            }
        assert model.dict() == {"id": "foo", "type": "exclude_model", "title": "foo", "foo": "long-random-thing"}


@pytest.fixture
def page_pre_defined_actions():
    return vm.Page(
        title="Page 1",
        components=[
            vm.Graph(figure=px.bar("iris", x="sepal_width", y="sepal_length")),
            vm.Button(
                text="Export data",
                actions=[
                    vm.Action(function=export_data()),
                    vm.Action(function=export_data()),
                ],
            ),
        ],
    )


@pytest.fixture
def page_two_captured_callables():
    @capture("graph")
    def chart(data_frame, hover_data: Optional[List[str]] = None):
        return px.bar(data_frame, x="sepal_width", y="sepal_length", hover_data=hover_data)

    @capture("graph")
    def chart2(data_frame, hover_data: Optional[List[str]] = None):
        return px.bar(data_frame, x="sepal_width", y="sepal_length", hover_data=hover_data)

    return vm.Page(
        title="Page 1",
        components=[
            vm.Graph(figure=chart(data_frame="iris")),
            vm.Graph(figure=chart2(data_frame="iris")),
        ],
    )


@pytest.fixture
def chart_dynamic():
    function_string = textwrap.dedent(
        """
        @capture("graph")
        def chart_dynamic(data_frame, hover_data: Optional[List[str]] = None):
            return px.bar(data_frame, x="sepal_width", y="sepal_length", hover_data=hover_data)
        """
    )

    exec(function_string)
    return locals()["chart_dynamic"]


@pytest.fixture
def complete_dashboard():
    @capture("graph")
    def chart(data_frame, hover_data: Optional[List[str]] = None):
        return px.bar(data_frame, x="sepal_width", y="sepal_length", hover_data=hover_data)

    page = vm.Page(
        title="Page 1",
        layout=vm.Layout(grid=[[0, 1], [2, 3], [4, -1]], row_min_height="100px"),
        components=[
            vm.Card(text="Foo"),
            vm.Graph(figure=px.bar("iris", x="sepal_width", y="sepal_length")),
            vm.Graph(figure=chart(data_frame="iris")),
            vm.AgGrid(figure=dash_ag_grid(data_frame="iris")),
            vm.Button(
                text="Export data",
                actions=[
                    vm.Action(function=export_data()),
                    vm.Action(function=export_data()),
                ],
            ),
        ],
        controls=[vm.Filter(column="species")],
    )

    dashboard = vm.Dashboard(title="Bar", pages=[page])

    return dashboard


expected_card = """############ Imports ##############
import vizro.models as vm


########### Model code ############
model = vm.Card(text="Foo")
"""


expected_graph = """############ Imports ##############
import vizro.plotly.express as px
import vizro.models as vm


####### Data Manager Settings #####
#######!!! UNCOMMENT BELOW !!!#####
# from vizro.managers import data_manager
# data_manager["iris"] = ===> Fill in here <===


########### Model code ############
model = vm.Graph(figure=px.bar(data_frame="iris", x="sepal_width", y="sepal_length"))
"""


expected_graph_with_callable = """############ Imports ##############
import vizro.plotly.express as px
import vizro.models as vm
from vizro.models.types import capture
from typing import Optional, List


####### Function definitions ######
@capture("graph")
def chart(data_frame, hover_data: Optional[List[str]] = None):
    return px.bar(data_frame, x="sepal_width", y="sepal_length", hover_data=hover_data)


####### Data Manager Settings #####
#######!!! UNCOMMENT BELOW !!!#####
# from vizro.managers import data_manager
# data_manager["iris"] = ===> Fill in here <===


########### Model code ############
model = vm.Graph(figure=chart(data_frame="iris"))
"""


expected_actions_predefined = """############ Imports ##############
import vizro.plotly.express as px
import vizro.models as vm
import vizro.actions as va


####### Data Manager Settings #####
#######!!! UNCOMMENT BELOW !!!#####
# from vizro.managers import data_manager
# data_manager["iris"] = ===> Fill in here <===


########### Model code ############
model = vm.Page(
    components=[
        vm.Graph(figure=px.bar(data_frame="iris", x="sepal_width", y="sepal_length")),
        vm.Button(
            text="Export data",
            actions=[
                vm.Action(function=va.export_data()),
                vm.Action(function=va.export_data()),
            ],
        ),
    ],
    title="Page 1",
)
"""


excepted_graph_dynamic = """############ Imports ##############
import vizro.models as vm


####### Data Manager Settings #####
#######!!! UNCOMMENT BELOW !!!#####
# from vizro.managers import data_manager
# data_manager["iris"] = ===> Fill in here <===


########### Model code ############
model = vm.Graph(figure=chart_dynamic(data_frame="iris"))
"""


extra_callable = """@capture("graph")
def extra(data_frame, hover_data: Optional[List[str]] = None):
    return px.bar(data_frame, x="sepal_width", y="sepal_length", hover_data=hover_data)
"""


expected_code_with_extra_callable = """############ Imports ##############
import vizro.plotly.express as px
import vizro.models as vm
from vizro.models.types import capture


####### Function definitions ######
@capture("graph")
def extra(data_frame, hover_data: Optional[List[str]] = None):
    return px.bar(data_frame, x="sepal_width", y="sepal_length", hover_data=hover_data)


########### Model code ############
model = vm.Card(text="Foo")
"""

expected_complete_dashboard = """############ Imports ##############
import vizro.plotly.express as px
import vizro.tables as vt
import vizro.models as vm
import vizro.actions as va
from vizro.models.types import capture
from typing import Optional, List


####### Function definitions ######
@capture("graph")
def chart(data_frame, hover_data: Optional[List[str]] = None):
    return px.bar(data_frame, x="sepal_width", y="sepal_length", hover_data=hover_data)


####### Data Manager Settings #####
#######!!! UNCOMMENT BELOW !!!#####
# from vizro.managers import data_manager
# data_manager["iris"] = ===> Fill in here <===


########### Model code ############
model = vm.Dashboard(
    pages=[
        vm.Page(
            components=[
                vm.Card(text="Foo"),
                vm.Graph(
                    figure=px.bar(data_frame="iris", x="sepal_width", y="sepal_length")
                ),
                vm.Graph(figure=chart(data_frame="iris")),
                vm.AgGrid(figure=vt.dash_ag_grid(data_frame="iris")),
                vm.Button(
                    text="Export data",
                    actions=[
                        vm.Action(function=va.export_data()),
                        vm.Action(function=va.export_data()),
                    ],
                ),
            ],
            title="Page 1",
            layout=vm.Layout(grid=[[0, 1], [2, 3], [4, -1]], row_min_height="100px"),
            controls=[vm.Filter(column="species")],
        )
    ],
    title="Bar",
)
"""


class TestPydanticPython:
    def test_to_python_basic(self):
        card = vm.Card(text="Foo")
        result = card._to_python()
        assert result == expected_card

    def test_to_python_data_manager(self):
        # Test if data manager code lines are included correctly in output
        graph = vm.Graph(figure=px.bar("iris", x="sepal_width", y="sepal_length"))
        result = graph._to_python()
        assert result == expected_graph

    def test_to_python_captured_callable_chart_plus_extra_imports(self):
        # Test if captured callable is included correctly in output
        # Test if extra imports are included correctly in output (typing yes, pandas no)
        @capture("graph")
        def chart(data_frame, hover_data: Optional[List[str]] = None):
            return px.bar(data_frame, x="sepal_width", y="sepal_length", hover_data=hover_data)

        graph = vm.Graph(figure=chart(data_frame="iris"))
        result = graph._to_python(extra_imports={"from typing import Optional, List", "import pandas as pd"})
        assert result == expected_graph_with_callable

    def test_to_python_two_captured_callable_charts(self, page_two_captured_callables):
        # Test if two captured callables are included. Note that the order in which they are included is not guaranteed.
        result = page_two_captured_callables._to_python()
        assert "def chart(data_frame, hover_data: Optional[List[str]] = None):" in result
        assert "def chart2(data_frame, hover_data: Optional[List[str]] = None):" in result

    def test_to_python_pre_defined_actions(self, page_pre_defined_actions):
        # Test if pre-defined actions are included correctly in output, ie no ActionsChain model
        result = page_pre_defined_actions._to_python()
        assert result == expected_actions_predefined

    def test_to_python_no_source_code(self, chart_dynamic, caplog):
        # Check if to_python works if the source code is not available - here chart_dynamic is undefined
        caplog.set_level(logging.INFO)

        graph = vm.Graph(figure=chart_dynamic(data_frame="iris"))
        result = graph._to_python()
        assert "Could not extract" in caplog.records[0].message
        assert result == excepted_graph_dynamic

    def test_to_python_with_extra_callable(self):
        # Test if callable is included correctly in output
        card = vm.Card(text="Foo")
        result = card._to_python(extra_callable_defs={extra_callable})
        assert result == expected_code_with_extra_callable

    def test_to_python_complete_dashboard(self, complete_dashboard):
        # Test more complete and nested model
        result = complete_dashboard._to_python(extra_imports={"from typing import Optional,List"})
        assert result == expected_complete_dashboard



import vizro.models as vm
import vizro.plotly.express as px
from vizro import Vizro

tips = px.data.tips()


page = vm.Page(
    title="Paired column",
    components=[
        vm.Graph(
            figure=px.histogram(
                tips,
                x="day",
                y="total_bill",
                color="sex",
                barmode="group",
                category_orders={"day": ["Thur", "Fri", "Sat", "Sun"]},
            ),
        )
    ],
)

dashboard = vm.Dashboard(pages=[page])
Vizro().build(dashboard).run()



"""Unit tests for vizro.models.Container."""

import dash_bootstrap_components as dbc
import pytest
from asserts import STRIP_ALL, assert_component_equal
from dash import html

try:
    from pydantic.v1 import ValidationError
except ImportError:  # pragma: no cov
    from pydantic import ValidationError

import vizro.models as vm


class TestContainerInstantiation:
    """Tests model instantiation and the validators run at that time."""

    def test_create_container_mandatory_only(self):
        container = vm.Container(title="Title", components=[vm.Button(), vm.Button()])
        assert isinstance(container.components[0], vm.Button) and isinstance(container.components[1], vm.Button)
        assert container.layout.grid == [[0], [1]]
        assert container.title == "Title"

    def test_create_container_mandatory_and_optional(self):
        container = vm.Container(
            id="my-id", title="Title", components=[vm.Button(), vm.Button()], layout=vm.Layout(grid=[[0, 1]])
        )
        assert container.id == "my-id"
        assert isinstance(container.components[0], vm.Button) and isinstance(container.components[1], vm.Button)
        assert container.layout.grid == [[0, 1]]
        assert container.title == "Title"

    def test_mandatory_title_missing(self):
        with pytest.raises(ValidationError, match="field required"):
            vm.Container(components=[vm.Button()])

    def test_mandatory_components_missing(self):
        with pytest.raises(ValidationError, match="field required"):
            vm.Container(title="Title")


class TestContainerBuildMethod:
    def test_container_build(self):
        result = vm.Container(
            id="container", title="Title", components=[vm.Button()], layout=vm.Layout(id="layout_id", grid=[[0]])
        ).build()
        assert_component_equal(
            result, html.Div(id="container", className="page-component-container"), keys_to_strip={"children"}
        )
        assert_component_equal(result.children, [html.H3(), html.Div()], keys_to_strip=STRIP_ALL)
        # We still want to test the exact H3 produced in Container.build:
        assert_component_equal(result.children[0], html.H3("Title", className="container__title"))
        # And also that a button has been inserted in the right place:
        assert_component_equal(result["layout_id_0"].children, dbc.Button(), keys_to_strip=STRIP_ALL)



"""Unit tests for DatePicker."""

from datetime import date, datetime

import dash_bootstrap_components as dbc
import dash_mantine_components as dmc
import pytest
from asserts import assert_component_equal
from dash import dcc, html

try:
    from pydantic.v1 import ValidationError
except ImportError:  # pragma: no cov
    from pydantic import ValidationError

import vizro.models as vm


class TestDatePickerInstantiation:
    """Tests model instantiation."""

    def test_create_datepicker_mandatory(self):
        date_picker = vm.DatePicker()

        assert hasattr(date_picker, "id")
        assert date_picker.type == "date_picker"
        assert date_picker.min is None
        assert date_picker.max is None
        assert date_picker.value is None
        assert date_picker.title == ""
        assert date_picker.actions == []
        assert date_picker.range is True

    def test_create_datepicker_mandatory_and_optional(self):
        date_picker = vm.DatePicker(
            id="date-picker-id", min="2024-01-01", max="2024-12-31", value=["2024-03-01", "2024-04-01"], title="Title"
        )

        assert date_picker.id == "date-picker-id"
        assert date_picker.type == "date_picker"
        assert date_picker.min == date(2024, 1, 1)
        assert date_picker.max == date(2024, 12, 31)
        assert date_picker.value == [date(2024, 3, 1), date(2024, 4, 1)]
        assert date_picker.title == "Title"
        assert date_picker.actions == []
        assert date_picker.range is True

    @pytest.mark.parametrize("title", ["test", 1, 1.0, """## Test header""", ""])
    def test_valid_title(self, title):
        date_picker = vm.DatePicker(title=title)

        assert date_picker.title == str(title)

    def test_set_action_via_validator(self, identity_action_function):
        date_picker = vm.DatePicker(actions=[vm.Action(function=identity_action_function())])
        actions_chain = date_picker.actions[0]

        assert actions_chain.trigger.component_property == "value"

    @pytest.mark.parametrize("min, max", [("2024-01-01", None), (None, "2024-01-01"), ("2024-01-01", "2024-02-01")])
    def test_valid_min_max(self, min, max):
        date_picker = vm.DatePicker(min=min, max=max)

        assert date_picker.min == (datetime.strptime(min, "%Y-%m-%d").date() if min else None)
        assert date_picker.max == (datetime.strptime(max, "%Y-%m-%d").date() if max else None)

    def test_validate_max_invalid_min_greater_than_max(self):
        with pytest.raises(
            ValidationError, match="Maximum value of selector is required to be larger than minimum value."
        ):
            vm.DatePicker(min="2024-02-01", max="2024-01-01")

    def test_validate_max_invalid_date_format(self):
        with pytest.raises(ValidationError, match="invalid date format"):
            vm.DatePicker(min="50-50-50", max="50-50-50")

    def test_validate_range_true_datepicker_value_valid(self):
        value = ["2024-01-01", "2024-02-01"]
        date_picker = vm.DatePicker(range=True, value=value)
        expected_value = [date(2024, 1, 1), date(2024, 2, 1)]
        assert date_picker.value == expected_value

    def test_validate_range_false_datepicker_value_valid(self):
        value = "2024-01-01"
        date_picker = vm.DatePicker(range=False, value=value)
        expected_value = date(2024, 1, 1)
        assert date_picker.value == expected_value

    @pytest.mark.parametrize("range, value", [(False, "2024-01-01"), (True, ["2024-01-01", "2024-02-01"])])
    def test_validate_datepicker_value_invalid(self, range, value):
        with pytest.raises(ValidationError, match="Please provide a valid value between the min and max value."):
            vm.DatePicker(min="1999-01-01", max="1999-02-01", range=range, value=value)

    @pytest.mark.parametrize("range, value", [(False, "2024-01-01"), (True, ["2024-01-01", "2024-02-01"])])
    def test_validate_datepicker_range_valid(self, range, value):
        date_picker = vm.DatePicker(min="2024-01-01", max="2024-02-01", range=range, value=value)
        assert date_picker.range == range

    @pytest.mark.parametrize(
        "range, value",
        [
            (True, "2024-01-01"),  # range True produces DateRangePicker, value needs to be list with 2 dates
            (True, ["2024-01-01"]),  # range True produces DateRangePicker, value needs to be list with 2 dates
            (False, ["2024-01-01"]),  # range False produces DatePicker, value needs to be single date
            (False, ["2024-01-01", "2024-03-01"]),  # range False produces DatePicker, value needs to be single date
        ],
    )
    def test_validate_datepicker_range_invalid(self, range, value):
        with pytest.raises(ValidationError):
            vm.DatePicker(range=range, value=value)


class TestBuildMethod:
    @pytest.mark.parametrize("range, value", [(False, "2023-01-05"), (True, ["2023-01-05", "2023-01-07"])])
    def test_datepicker_build(self, range, value):
        date_picker = vm.DatePicker(
            min="2023-01-01", max="2023-07-01", range=range, value=value, id="datepicker_id", title="Test title"
        ).build()

        date_picker_class = dmc.DateRangePicker if range else dmc.DatePicker
        additional_kwargs = {"allowSingleDateInRange": True} if range else {}
        expected_datepicker = html.Div(
            [
                dbc.Label("Test title", html_for="datepicker_id"),
                date_picker_class(
                    id="datepicker_id",
                    minDate="2023-01-01",
                    maxDate="2023-07-02",
                    value=value,
                    persistence=True,
                    persistence_type="session",
                    dropdownPosition="bottom-start",
                    disabledDates="2023-07-02",
                    clearable=False,
                    className="datepicker",
                    **additional_kwargs,
                ),
                dcc.Store(id="datepicker_id_input_store", storage_type="session", data=value),
            ],
            className="selector_container",
        )
        assert_component_equal(date_picker, expected_datepicker)



import vizro.models as vm
import vizro.plotly.express as px
from vizro import Vizro

gapminder = px.data.gapminder()

page = vm.Page(
    title="Bubble",
    components=[
        vm.Graph(figure=px.scatter(gapminder.query("year==2007"), x="gdpPercap", y="lifeExp", size="pop", size_max=60))
    ],
)

dashboard = vm.Dashboard(pages=[page])
Vizro().build(dashboard).run()



"""Unit tests for vizro.models.Card."""

import dash_bootstrap_components as dbc
import pytest
from asserts import assert_component_equal
from dash import dcc

try:
    from pydantic.v1 import ValidationError
except ImportError:  # pragma: no cov
    from pydantic import ValidationError

import vizro.models as vm


class TestCardInstantiation:
    """Tests model instantiation."""

    def test_create_card_mandatory_only(self):
        card = vm.Card(text="Text to test card")

        assert hasattr(card, "id")
        assert card.type == "card"
        assert card.text == "Text to test card"
        assert card.href == ""

    @pytest.mark.parametrize("id, href", [("id_1", "/page_1_reference"), ("id_2", "https://www.google.de/")])
    def test_create_card_mandatory_and_optional(self, id, href):
        card = vm.Card(id=id, text="Text to test card", href=href)

        assert card.id == id
        assert card.type == "card"
        assert card.text == "Text to test card"
        assert card.href == href

    def test_mandatory_text_missing(self):
        with pytest.raises(ValidationError, match="field required"):
            vm.Card()

    def test_none_as_text(self):
        with pytest.raises(ValidationError, match="none is not an allowed value"):
            vm.Card(text=None)


class TestBuildMethod:
    """Tests build method."""

    def test_card_build_with_href(self):
        card = vm.Card(id="card_id", text="Hello", href="https://www.google.com")
        card = card.build()

        expected_card = dbc.Card(
            dbc.NavLink(
                dcc.Markdown(id="card_id", children="Hello", dangerously_allow_html=False),
                href="https://www.google.com",
                target="_top",
            ),
            className="card-nav",
        )

        assert_component_equal(card, expected_card)

    def test_card_build_wo_href(self):
        card = vm.Card(id="card_id", text="Hello")
        card = card.build()
        assert_component_equal(
            card, dbc.Card(dcc.Markdown(id="card_id", children="Hello", dangerously_allow_html=False), className="")
        )

    @pytest.mark.parametrize(
        "test_text, expected",
        [
            ("""# Header""", "# Header"),
            ("""_This **tests** emphasis_""", "_This **tests** emphasis_"),
            ("""> This tests blockquotes""", "> This tests blockquotes"),
            ("""* This tests list items""", "* This tests list items"),
            ("Text to test card", "Text to test card"),
            ("", ""),
            (
                """![](assets/images/icons/content/hypotheses.svg#icon-top)""",
                "![](assets/images/icons/content/hypotheses.svg#icon-top)",
            ),
            ("""Code block: ```python print(1)```""", "Code block: ```python print(1)```"),
            ("""[Example page](/test_page)""", "[Example page](/test_page)"),
        ],
    )
    def test_markdown_setting(self, test_text, expected):
        card = vm.Card(id="id_valid", text=test_text)
        card = card.build()
        card_markdown = card["id_valid"]

        assert isinstance(card_markdown, dcc.Markdown)
        assert card_markdown.dangerously_allow_html is False
        assert card_markdown.children == expected

    @pytest.mark.parametrize(
        "test_text, expected",
        [
            ("""<p>Hello </p>""", "<p>Hello </p>"),  # html will not be evaluated but converted to string
            (12345, "12345"),
            ("""$$ \\frac{1}{(\\sqrt{\\phi \\sqrt{5}}-\\phi)}$$""", "$$ \\frac{1}{(\\sqrt{\\phi \\sqrt{5}}-\\phi)}$$"),
        ],
    )
    def test_markdown_build_invalid(self, test_text, expected):
        card = vm.Card(id="test_id", text=test_text)
        card = card.build()
        card_markdown = card["test_id"]

        assert isinstance(card_markdown, dcc.Markdown)
        assert card_markdown.dangerously_allow_html is False
        assert card_markdown.children == expected



from __future__ import annotations

from typing import Any, List, Mapping, Optional, Set, TypedDict, Union

from dash import dcc, html

try:
    from pydantic.v1 import Field, root_validator, validator
except ImportError:  # pragma: no cov
    from pydantic import Field, root_validator, validator

from vizro._constants import ON_PAGE_LOAD_ACTION_PREFIX
from vizro.actions import _on_page_load
from vizro.managers import model_manager
from vizro.managers._model_manager import DuplicateIDError, ModelID
from vizro.models import Action, Layout, VizroBaseModel
from vizro.models._action._actions_chain import ActionsChain, Trigger
from vizro.models._layout import set_layout
from vizro.models._models_utils import _log_call, check_captured_callable, validate_min_length

from .types import ComponentType, ControlType

# This is just used for type checking. Ideally it would inherit from some dash.development.base_component.Component
# (e.g. html.Div) as well as TypedDict, but that's not possible, and Dash does not have typing support anyway. When
# this type is used, the object is actually still a dash.development.base_component.Component, but this makes it easier
# to see what contract the component fulfills by making the expected keys explicit.
_PageBuildType = TypedDict("_PageBuildType", {"control-panel": html.Div, "page-components": html.Div})


class Page(VizroBaseModel):
    """A page in [`Dashboard`][vizro.models.Dashboard] with its own URL path and place in the `Navigation`.

    Args:
        components (List[ComponentType]): See [ComponentType][vizro.models.types.ComponentType]. At least one component
            has to be provided.
        title (str): Title to be displayed.
        description (str): Description for meta tags.
        layout (Layout): Layout to place components in. Defaults to `None`.
        controls (List[ControlType]): See [ControlType][vizro.models.types.ControlType]. Defaults to `[]`.
        path (str): Path to navigate to page. Defaults to `""`.

    """

    components: List[ComponentType]
    title: str = Field(..., description="Title to be displayed.")
    description: str = Field("", description="Description for meta tags.")
    layout: Layout = None  # type: ignore[assignment]
    controls: List[ControlType] = []
    path: str = Field("", description="Path to navigate to page.")

    # TODO: Remove default on page load action if possible
    actions: List[ActionsChain] = []

    # Re-used validators
    _check_captured_callable = validator("components", allow_reuse=True, each_item=True, pre=True)(
        check_captured_callable
    )
    _validate_min_length = validator("components", allow_reuse=True, always=True)(validate_min_length)
    _validate_layout = validator("layout", allow_reuse=True, always=True)(set_layout)

    @root_validator(pre=True)
    def set_id(cls, values):
        if "title" not in values:
            return values

        values.setdefault("id", values["title"])
        return values

    @validator("path", always=True)
    def set_path(cls, path, values) -> str:
        # Based on how Github generates anchor links - see:
        # https://stackoverflow.com/questions/72536973/how-are-github-markdown-anchor-links-constructed.
        def clean_path(path: str, allowed_characters: str) -> str:
            path = path.strip().lower().replace(" ", "-")
            path = "".join(character for character in path if character.isalnum() or character in allowed_characters)
            return path if path.startswith("/") else "/" + path

        # Allow "/" in path if provided by user, otherwise turn page id into suitable URL path (not allowing "/")
        if path:
            return clean_path(path, "-_/")
        return clean_path(values["id"], "-_")

    def __init__(self, **data):
        """Adds the model instance to the model manager."""
        try:
            super().__init__(**data)
        except DuplicateIDError as exc:
            raise ValueError(
                f"Page with id={self.id} already exists. Page id is automatically set to the same "
                f"as the page title. If you have multiple pages with the same title then you must assign a unique id."
            ) from exc

    def __vizro_exclude_fields__(self) -> Optional[Union[Set[str], Mapping[str, Any]]]:
        return {"id"} if self.id == self.title else None

    @_log_call
    def pre_build(self):
        # TODO: Remove default on page load action if possible
        targets = model_manager._get_page_model_ids_with_figure(page_id=ModelID(str(self.id)))
        if targets:
            self.actions = [
                ActionsChain(
                    id=f"{ON_PAGE_LOAD_ACTION_PREFIX}_{self.id}",
                    trigger=Trigger(
                        component_id=f"{ON_PAGE_LOAD_ACTION_PREFIX}_trigger_{self.id}", component_property="data"
                    ),
                    actions=[
                        Action(
                            id=f"{ON_PAGE_LOAD_ACTION_PREFIX}_action_{self.id}", function=_on_page_load(targets=targets)
                        )
                    ],
                )
            ]

    @_log_call
    def build(self) -> _PageBuildType:
        controls_content = [control.build() for control in self.controls]
        control_panel = html.Div(id="control-panel", children=controls_content, hidden=not controls_content)

        components_container = self.layout.build()
        for component_idx, component in enumerate(self.components):
            components_container[f"{self.layout.id}_{component_idx}"].children = component.build()

        # Page specific CSS ID and Stores
        components_container.children.append(dcc.Store(id=f"{ON_PAGE_LOAD_ACTION_PREFIX}_trigger_{self.id}"))
        components_container.id = "page-components"
        return html.Div([control_panel, components_container])



import vizro.models as vm
import vizro.plotly.express as px
from vizro import Vizro

tips = px.data.tips()

page = vm.Page(
    title="Donut",
    components=[vm.Graph(figure=px.pie(tips, values="tip", names="day", hole=0.4))],
)

dashboard = vm.Dashboard(pages=[page])
Vizro().build(dashboard).run()



"""Contains utilities for the implementation of action functions."""

from __future__ import annotations

from collections import defaultdict
from copy import deepcopy
from typing import TYPE_CHECKING, Any, Dict, List, Literal, Optional, TypedDict, Union

import pandas as pd

from vizro._constants import ALL_OPTION, NONE_OPTION
from vizro.managers import data_manager, model_manager
from vizro.managers._model_manager import ModelID
from vizro.models.types import MultiValueType, SelectorType, SingleValueType

if TYPE_CHECKING:
    from vizro.models import Action, VizroBaseModel

ValidatedNoneValueType = Union[SingleValueType, MultiValueType, None, List[None]]


class CallbackTriggerDict(TypedDict):
    """Represent dash.ctx.args_grouping item. Shortened as 'ctd' in the code.

    Args:
        id: The component ID. If it`s a pattern matching ID, it will be a dict.
        property: The component property used in the callback.
        value: The value of the component property at the time the callback was fired.
        str_id: For pattern matching IDs, it's the stringified dict ID without white spaces.
        triggered: A boolean indicating whether this input triggered the callback.

    """

    id: ModelID
    property: Literal["clickData", "value", "n_clicks", "active_cell", "derived_viewport_data"]
    value: Optional[Any]
    str_id: str
    triggered: bool


# Utility functions for helper functions used in pre-defined actions ----
def _get_component_actions(component) -> List[Action]:
    return (
        [action for actions_chain in component.actions for action in actions_chain.actions]
        if hasattr(component, "actions")
        else []
    )


def _apply_filters(data_frame: pd.DataFrame, ctds_filters: List[CallbackTriggerDict], target: str) -> pd.DataFrame:
    for ctd in ctds_filters:
        selector_value = ctd["value"]
        selector_value = selector_value if isinstance(selector_value, list) else [selector_value]
        selector_actions = _get_component_actions(model_manager[ctd["id"]])

        for action in selector_actions:
            if (
                action.function._function.__name__ != "_filter"
                or target not in action.function["targets"]
                or ALL_OPTION in selector_value
            ):
                continue

            _filter_function = action.function["filter_function"]
            _filter_column = action.function["filter_column"]
            _filter_value = selector_value
            data_frame = data_frame[_filter_function(data_frame[_filter_column], _filter_value)]

    return data_frame


def _get_parent_vizro_model(_underlying_callable_object_id: str) -> VizroBaseModel:
    from vizro.models import VizroBaseModel

    for _, vizro_base_model in model_manager._items_with_type(VizroBaseModel):
        if (
            hasattr(vizro_base_model, "_input_component_id")
            and vizro_base_model._input_component_id == _underlying_callable_object_id
        ):
            return vizro_base_model
    raise KeyError(
        f"No parent Vizro model found for underlying callable object with id: {_underlying_callable_object_id}."
    )


def _apply_filter_interaction(
    data_frame: pd.DataFrame, ctds_filter_interaction: List[Dict[str, CallbackTriggerDict]], target: str
) -> pd.DataFrame:
    for ctd_filter_interaction in ctds_filter_interaction:
        triggered_model = model_manager[ctd_filter_interaction["modelID"]["id"]]
        data_frame = triggered_model._filter_interaction(
            data_frame=data_frame,
            target=target,
            ctd_filter_interaction=ctd_filter_interaction,
        )

    return data_frame


def _validate_selector_value_none(value: Union[SingleValueType, MultiValueType]) -> ValidatedNoneValueType:
    if value == NONE_OPTION:
        return None
    elif isinstance(value, list):
        return [i for i in value if i != NONE_OPTION] or [None]
    return value


def _create_target_arg_mapping(dot_separated_strings: List[str]) -> Dict[str, List[str]]:
    results = defaultdict(list)
    for string in dot_separated_strings:
        if "." not in string:
            raise ValueError(f"Provided string {string} must contain a '.'")
        component, arg = string.split(".", 1)
        results[component].append(arg)
    return results


def _update_nested_graph_properties(
    graph_config: Dict[str, Any], dot_separated_string: str, value: Any
) -> Dict[str, Any]:
    keys = dot_separated_string.split(".")
    current_property = graph_config

    for key in keys[:-1]:
        current_property = current_property.setdefault(key, {})

    current_property[keys[-1]] = value
    return graph_config


def _get_parametrized_config(target: ModelID, ctd_parameters: List[CallbackTriggerDict]) -> Dict[str, Any]:
    # TODO - avoid calling _captured_callable. Once we have done this we can remove _arguments from
    #  CapturedCallable entirely.
    config = deepcopy(model_manager[target].figure._arguments)

    # It's not possible to address nested argument of data_frame like data_frame.x.y, just top-level ones like
    # data_frame.x.
    config["data_frame"] = {}

    for ctd in ctd_parameters:
        # TODO: needs to be refactored so that it is independent of implementation details
        selector_value = ctd["value"]

        if hasattr(selector_value, "__iter__") and ALL_OPTION in selector_value:  # type: ignore[operator]
            selector: SelectorType = model_manager[ctd["id"]]

            # Even if options are provided as List[Dict], the Dash component only returns a List of values.
            # So we need to ensure that we always return a List only as well to provide consistent types.
            if all(isinstance(option, dict) for option in selector.options):
                selector_value = [option["value"] for option in selector.options]
            else:
                selector_value = selector.options

        selector_value = _validate_selector_value_none(selector_value)
        selector_actions = _get_component_actions(model_manager[ctd["id"]])

        for action in selector_actions:
            if action.function._function.__name__ != "_parameter":
                continue

            action_targets = _create_target_arg_mapping(action.function["targets"])

            if target not in action_targets:
                continue

            for action_targets_arg in action_targets[target]:
                config = _update_nested_graph_properties(
                    graph_config=config, dot_separated_string=action_targets_arg, value=selector_value
                )

    return config


# Helper functions used in pre-defined actions ----
def _get_targets_data_and_config(
    ctds_filter: List[CallbackTriggerDict],
    ctds_filter_interaction: List[Dict[str, CallbackTriggerDict]],
    ctds_parameters: List[CallbackTriggerDict],
    targets: List[ModelID],
):
    all_filtered_data = {}
    all_parameterized_config = {}

    for target in targets:
        # parametrized_config includes a key "data_frame" that is used in the data loading function.
        parameterized_config = _get_parametrized_config(target=target, ctd_parameters=ctds_parameters)
        data_source_name = model_manager[target]["data_frame"]
        data_frame = data_manager[data_source_name].load(**parameterized_config["data_frame"])

        filtered_data = _apply_filters(data_frame=data_frame, ctds_filters=ctds_filter, target=target)
        filtered_data = _apply_filter_interaction(
            data_frame=filtered_data, ctds_filter_interaction=ctds_filter_interaction, target=target
        )

        # Parameters affecting data_frame have already been used above in data loading and so are excluded from
        # all_parameterized_config.
        all_filtered_data[target] = filtered_data
        all_parameterized_config[target] = {
            key: value for key, value in parameterized_config.items() if key != "data_frame"
        }

    return all_filtered_data, all_parameterized_config


def _get_modified_page_figures(
    ctds_filter: List[CallbackTriggerDict],
    ctds_filter_interaction: List[Dict[str, CallbackTriggerDict]],
    ctds_parameters: List[CallbackTriggerDict],
    targets: Optional[List[ModelID]] = None,
) -> Dict[str, Any]:
    targets = targets or []

    filtered_data, parameterized_config = _get_targets_data_and_config(
        ctds_filter=ctds_filter,
        ctds_filter_interaction=ctds_filter_interaction,
        ctds_parameters=ctds_parameters,
        targets=targets,
    )

    outputs: Dict[str, Any] = {}
    for target in targets:
        outputs[target] = model_manager[target](data_frame=filtered_data[target], **parameterized_config[target])

    return outputs



import re
from datetime import date, datetime
from typing import Literal

import pandas as pd
import pytest
import vizro.models as vm
from asserts import assert_component_equal
from vizro.managers import model_manager
from vizro.models._action._actions_chain import ActionsChain
from vizro.models._controls.filter import Filter, _filter_between, _filter_isin
from vizro.models.types import CapturedCallable


class TestFilterFunctions:
    @pytest.mark.parametrize(
        "data, value, expected",
        [
            ([1, 2, 3, 4, 5], [2.5, 3.5], [False, False, True, False, False]),  # Standard test
            ([1, 2, 3, 4, 5], [2, 4], [False, True, True, True, False]),  # Test for inclusive both ends
            ([1, 2, 3, 4, 5], [1, 5], [True, True, True, True, True]),  # Test for inclusive all
            ([1, 2, 3, 4, 5], [4, 2], [False, False, False, False, False]),  # Test for inverted values
            ([], [2, 4], pd.Series([], dtype=bool)),  # Test for empty series
            ([1.1, 2.2, 3.3, 4.4, 5.5], [2.1, 4.5], [False, True, True, True, False]),  # Test with float data
        ],
    )
    def test_filter_between(self, data, value, expected):
        series = pd.Series(data)
        expected = pd.Series(expected)
        result = _filter_between(series, value)
        pd.testing.assert_series_equal(result, expected)

    @pytest.mark.parametrize(
        "data, value, expected",
        [
            (
                [
                    datetime(2024, 1, 1),
                    datetime(2024, 2, 1),
                    datetime(2024, 3, 1),
                    datetime(2024, 4, 1),
                    datetime(2024, 5, 1),
                ],
                ["2024-02-01", "2024-03-01"],
                [False, True, True, False, False],
            ),  # Standard test
            (
                [
                    datetime(2024, 1, 1),
                    datetime(2024, 2, 1),
                    datetime(2024, 3, 1),
                    datetime(2024, 4, 1),
                    datetime(2024, 5, 1),
                ],
                ["2024-01-01", "2024-05-01"],
                [True, True, True, True, True],
            ),  # Test with dates for inclusive both ends
            (
                [
                    datetime(2024, 1, 1),
                    datetime(2024, 2, 1),
                    datetime(2024, 3, 1),
                    datetime(2024, 4, 1),
                    datetime(2024, 5, 1),
                ],
                ["2024-06-01", "2024-07-01"],
                [False, False, False, False, False],
            ),  # Test with no result
            (
                [
                    datetime(2024, 1, 1),
                    datetime(2024, 2, 1),
                    datetime(2024, 3, 1),
                    datetime(2024, 4, 1),
                    datetime(2024, 5, 1),
                ],
                ["2024-03-01", "2024-02-01"],
                [False, False, False, False, False],
            ),  # Test for inverted values
            ([], ["2024-02-01", "2024-03-01"], pd.Series([], dtype=bool)),  # Test for empty series
            (
                [
                    datetime(2024, 1, 1, 20, 20, 20),
                    datetime(2024, 2, 1, 20, 20, 20),
                    datetime(2024, 3, 1, 20, 20, 20),
                    datetime(2024, 4, 1, 20, 20, 20),
                    datetime(2024, 5, 1, 20, 20, 20),
                ],
                ["2024-02-01", "2024-03-01"],
                [False, True, True, False, False],
            ),  # Test with time part in the date
        ],
    )
    def test_filter_between_date(self, data, value, expected):
        series = pd.Series(data)
        expected = pd.Series(expected)
        result = _filter_between(series, value)
        pd.testing.assert_series_equal(result, expected)

    @pytest.mark.parametrize(
        "data, value, expected",
        [
            ([1, 2, 3, 4, 5], [2, 4], [False, True, False, True, False]),  # Test for integers
            (["apple", "banana", "orange"], ["banana", "grape"], [False, True, False]),  # Test for strings
            ([1.1, 2.2, 3.3, 4.4, 5.5], [2.2, 4.4], [False, True, False, True, False]),  # Test for float values
            ([1, 2, 3, 4, 5], [], [False, False, False, False, False]),  # Test for empty value list
        ],
    )
    def test_filter_isin(self, data, value, expected):
        series = pd.Series(data)
        expected = pd.Series(expected)
        result = _filter_isin(series, value)
        pd.testing.assert_series_equal(result, expected)

    @pytest.mark.parametrize(
        "data, value, expected",
        [
            (
                [
                    datetime(2024, 1, 1),
                    datetime(2024, 2, 1),
                    datetime(2024, 3, 1),
                    datetime(2024, 4, 1),
                    datetime(2024, 5, 1),
                ],
                ["2024-02-01"],
                [False, True, False, False, False],
            ),  # Standard test
            (
                [
                    datetime(2024, 1, 1),
                    datetime(2024, 2, 1),
                    datetime(2024, 3, 1),
                    datetime(2024, 2, 1),
                    datetime(2024, 2, 1),
                ],
                ["2024-02-01"],
                [False, True, False, True, True],
            ),  # Multiple values
            (
                [
                    datetime(2024, 1, 1),
                    datetime(2024, 2, 1),
                    datetime(2024, 3, 1),
                    datetime(2024, 4, 1),
                    datetime(2024, 5, 1),
                ],
                ["2024-06-01"],
                [False, False, False, False, False],
            ),  # Test with no result
            (
                [
                    datetime(2024, 1, 1),
                    datetime(2024, 2, 1),
                    datetime(2024, 3, 1),
                    datetime(2024, 4, 1),
                    datetime(2024, 5, 1),
                ],
                [],
                [False, False, False, False, False],
            ),  # Test for empty value list
            (
                [
                    datetime(2024, 1, 1, 20, 20, 20),
                    datetime(2024, 2, 1, 20, 20, 20),
                    datetime(2024, 3, 1, 20, 20, 20),
                    datetime(2024, 4, 1, 20, 20, 20),
                    datetime(2024, 5, 1, 20, 20, 20),
                ],
                ["2024-02-01"],
                [False, True, False, False, False],
            ),  # Test with time part in the date
        ],
    )
    def test_filter_isin_date(self, data, value, expected):
        series = pd.Series(data)
        expected = pd.Series(expected)
        result = _filter_isin(series, value)
        pd.testing.assert_series_equal(result, expected)


@pytest.mark.usefixtures("managers_one_page_two_graphs")
class TestFilterInstantiation:
    """Tests model instantiation and the validators run at that time."""

    def test_create_filter_mandatory_only(self):
        filter = Filter(column="foo")
        assert filter.type == "filter"
        assert filter.column == "foo"
        assert filter.targets == []

    def test_create_filter_mandatory_and_optional(self):
        filter = Filter(column="foo", targets=["scatter_chart", "bar_chart"], selector=vm.RadioItems())
        assert filter.type == "filter"
        assert filter.column == "foo"
        assert filter.targets == ["scatter_chart", "bar_chart"]
        assert isinstance(filter.selector, vm.RadioItems)

    def test_check_target_present_valid(self):
        Filter(column="foo", targets=["scatter_chart", "bar_chart"])

    def test_check_target_present_invalid(self):
        with pytest.raises(ValueError, match="Target invalid_target not found in model_manager."):
            Filter(column="foo", targets=["invalid_target"])


class TestPreBuildMethod:
    def test_set_targets_valid(self, managers_one_page_two_graphs):
        # Core of tests is still interface level
        filter = vm.Filter(column="country")
        # Special case - need filter in the context of page in order to run filter.pre_build
        model_manager["test_page"].controls = [filter]
        filter.pre_build()
        assert set(filter.targets) == {"scatter_chart", "bar_chart"}

    def test_set_targets_invalid(self, managers_one_page_two_graphs):
        filter = vm.Filter(column="invalid_choice")
        model_manager["test_page"].controls = [filter]

        with pytest.raises(ValueError, match="Selected column invalid_choice not found in any dataframe on this page."):
            filter.pre_build()

    @pytest.mark.parametrize(
        "filtered_column, expected_column_type",
        [("country", "categorical"), ("year", "temporal"), ("lifeExp", "numerical")],
    )
    def test_set_column_type(self, filtered_column, expected_column_type, managers_one_page_two_graphs):
        filter = vm.Filter(column=filtered_column)
        model_manager["test_page"].controls = [filter]
        filter.pre_build()
        assert filter._column_type == expected_column_type

    @pytest.mark.parametrize(
        "filtered_column, expected_selector",
        [("country", vm.Dropdown), ("year", vm.DatePicker), ("lifeExp", vm.RangeSlider)],
    )
    def test_set_selector_default_selector(self, filtered_column, expected_selector, managers_one_page_two_graphs):
        filter = vm.Filter(column=filtered_column)
        model_manager["test_page"].controls = [filter]
        filter.pre_build()
        assert isinstance(filter.selector, expected_selector)
        assert filter.selector.title == filtered_column.title()

    @pytest.mark.parametrize("filtered_column", ["country", "year", "lifeExp"])
    def test_set_selector_specific_selector(self, filtered_column, managers_one_page_two_graphs):
        filter = vm.Filter(column=filtered_column, selector=vm.RadioItems(title="Title"))
        model_manager["test_page"].controls = [filter]
        filter.pre_build()
        assert isinstance(filter.selector, vm.RadioItems)
        assert filter.selector.title == "Title"

    @pytest.mark.parametrize(
        "filtered_column, selector",
        [
            ("country", vm.Dropdown),
            ("country", vm.RadioItems),
            ("country", vm.Checklist),
            ("lifeExp", vm.Slider),
            ("lifeExp", vm.RangeSlider),
            ("lifeExp", vm.Dropdown),
            ("lifeExp", vm.RadioItems),
            ("lifeExp", vm.Checklist),
            ("year", vm.Dropdown),
            ("year", vm.RadioItems),
            ("year", vm.Checklist),
            ("year", vm.DatePicker),
        ],
    )
    def test_allowed_selectors_per_column_type(self, filtered_column, selector, managers_one_page_two_graphs):
        filter = vm.Filter(column=filtered_column, selector=selector())
        model_manager["test_page"].controls = [filter]
        filter.pre_build()
        assert isinstance(filter.selector, selector)

    @pytest.mark.parametrize(
        "filtered_column, selector",
        [
            ("country", vm.Slider),
            ("country", vm.RangeSlider),
            ("country", vm.DatePicker),
            ("lifeExp", vm.DatePicker),
            ("year", vm.Slider),
            ("year", vm.RangeSlider),
        ],
    )
    def test_disallowed_selectors_per_column_type(self, filtered_column, selector, managers_one_page_two_graphs):
        filter = vm.Filter(column=filtered_column, selector=selector())
        model_manager["test_page"].controls = [filter]
        with pytest.raises(
            ValueError,
            match=f"Chosen selector {selector().type} is not compatible with .* column '{filtered_column}'. ",
        ):
            filter.pre_build()

    @pytest.mark.parametrize(
        "targets",
        [
            ["id_shared_column_numerical", "id_shared_column_temporal"],
            ["id_shared_column_numerical", "id_shared_column_categorical"],
            ["id_shared_column_temporal", "id_shared_column_categorical"],
        ],
    )
    def test_set_slider_values_shared_column_inconsistent_dtype(self, targets, managers_shared_column_different_dtype):
        filter = vm.Filter(column="shared_column", targets=targets)
        model_manager["graphs_with_shared_column"].controls = [filter]
        with pytest.raises(
            ValueError,
            match=re.escape(
                f"Inconsistent types detected in the shared data column 'shared_column' for targeted charts {targets}. "
                f"Please ensure that the data column contains the same data type across all targeted charts."
            ),
        ):
            filter.pre_build()

    @pytest.mark.parametrize("selector", [vm.Slider, vm.RangeSlider])
    def test_set_numerical_selectors_values_min_max_default(self, selector, gapminder, managers_one_page_two_graphs):
        filter = vm.Filter(column="lifeExp", selector=selector())
        model_manager["test_page"].controls = [filter]
        filter.pre_build()
        assert filter.selector.min == gapminder.lifeExp.min()
        assert filter.selector.max == gapminder.lifeExp.max()

    def test_set_temporal_selectors_values_min_max_default(self, gapminder, managers_one_page_two_graphs):
        filter = vm.Filter(column="year", selector=vm.DatePicker())
        model_manager["test_page"].controls = [filter]
        filter.pre_build()
        assert filter.selector.min == gapminder.year.min().to_pydatetime().date()
        assert filter.selector.max == gapminder.year.max().to_pydatetime().date()

    @pytest.mark.parametrize("selector", [vm.Slider, vm.RangeSlider])
    def test_set_numerical_selectors_values_min_max_specific(self, selector, managers_one_page_two_graphs):
        filter = vm.Filter(column="lifeExp", selector=selector(min=3, max=5))
        model_manager["test_page"].controls = [filter]
        filter.pre_build()
        assert filter.selector.min == 3
        assert filter.selector.max == 5

    def test_set_temporal_selectors_values_min_max_specific(self, managers_one_page_two_graphs):
        filter = vm.Filter(column="year", selector=vm.DatePicker(min="1952-01-01", max="2007-01-01"))
        model_manager["test_page"].controls = [filter]
        filter.pre_build()
        assert filter.selector.min == date(1952, 1, 1)
        assert filter.selector.max == date(2007, 1, 1)

    @pytest.mark.parametrize("selector", [vm.Checklist, vm.Dropdown, vm.RadioItems])
    def test_set_categorical_selectors_options_default(self, selector, gapminder, managers_one_page_two_graphs):
        filter = vm.Filter(column="continent", selector=selector())
        model_manager["test_page"].controls = [filter]
        filter.pre_build()
        assert filter.selector.options == sorted(set(gapminder["continent"]))

    @pytest.mark.parametrize("selector", [vm.Checklist, vm.Dropdown, vm.RadioItems])
    def test_set_categorical_selectors_options_specific(self, selector, managers_one_page_two_graphs):
        filter = vm.Filter(column="continent", selector=selector(options=["Africa", "Europe"]))
        model_manager["test_page"].controls = [filter]
        filter.pre_build()
        assert filter.selector.options == ["Africa", "Europe"]

    @pytest.mark.parametrize(
        "filtered_column, selector, filter_function",
        [
            ("lifeExp", None, _filter_between),
            ("country", None, _filter_isin),
            ("year", None, _filter_between),
            ("year", vm.DatePicker(range=False), _filter_isin),
        ],
    )
    def test_set_actions(self, filtered_column, selector, filter_function, managers_one_page_two_graphs):
        filter = vm.Filter(column=filtered_column, selector=selector)
        model_manager["test_page"].controls = [filter]
        filter.pre_build()
        default_action = filter.selector.actions[0]
        assert isinstance(default_action, ActionsChain)
        assert isinstance(default_action.actions[0].function, CapturedCallable)
        assert default_action.actions[0].function["filter_function"] == filter_function
        assert default_action.actions[0].id == f"filter_action_{filter.id}"

    # TODO: Add tests for custom temporal and categorical selectors too. Probably inside the conftest file and reused in
    #       all other tests. Also add tests for the custom selector that is an entirely new component and adjust docs.
    # This test does add_type so ideally we would clean up after this to restore vizro.models to its previous state.
    # This is difficult to fix fully by un-importing vizro.models though, since we use `import vizro.models as vm` - see
    # https://stackoverflow.com/questions/437589/how-do-i-unload-reload-a-python-module.
    def test_numerical_custom_selector(self, gapminder, managers_one_page_two_graphs):
        class RangeSliderNonCross(vm.RangeSlider):
            """Custom numerical multi-selector `RangeSliderNonCross` to be provided to `Filter`."""

            type: Literal["range_slider_non_cross"] = "range_slider_non_cross"

            def build(self):
                range_slider_build_obj = super().build()
                range_slider_build_obj[self.id].allowCross = False
                return range_slider_build_obj

        filtered_column = "lifeExp"
        selector = RangeSliderNonCross
        vm.Filter.add_type("selector", selector)

        filter = vm.Filter(column=filtered_column, selector=selector())
        model_manager["test_page"].controls = [filter]
        filter.pre_build()

        assert isinstance(filter.selector, selector)
        assert filter.selector.title == filtered_column.title()
        assert filter.selector.min == gapminder.lifeExp.min()
        assert filter.selector.max == gapminder.lifeExp.max()

        default_action = filter.selector.actions[0]
        assert isinstance(default_action, ActionsChain)
        assert isinstance(default_action.actions[0].function, CapturedCallable)
        assert default_action.actions[0].function["filter_function"] == _filter_between
        assert default_action.actions[0].id == f"filter_action_{filter.id}"


@pytest.mark.usefixtures("managers_one_page_two_graphs")
class TestFilterBuild:
    """Tests filter build method."""

    @pytest.mark.parametrize(
        "test_column,test_selector",
        [
            ("continent", vm.Checklist()),
            ("continent", vm.Dropdown()),
            ("continent", vm.RadioItems()),
            ("pop", vm.RangeSlider()),
            ("pop", vm.Slider()),
            ("year", vm.DatePicker()),
            ("year", vm.DatePicker(range=False)),
        ],
    )
    def test_filter_build(self, test_column, test_selector):
        filter = vm.Filter(column=test_column, selector=test_selector)
        model_manager["test_page"].controls = [filter]
        filter.pre_build()
        result = filter.build()
        expected = test_selector.build()

        assert_component_equal(result, expected)



"""App configuration for dashboard."""

from typing import List, Union

import vizro.models as vm
from chart_groups import ALL_CHART_GROUP, CHART_GROUPS, ChartGroup, IncompletePage
from custom_components import FlexContainer, Markdown
from vizro import Vizro


def make_chart_card(page: Union[vm.Page, IncompletePage]) -> vm.Card:
    """Makes a card with svg icon, linked to the right page if page is complete.

    Args:
        page: page to make card for

    Returns: card with svg icon, linked to the right page if page is complete.

    """
    # There's one SVG per chart title, so that e.g. pages distribution-butterfly and deviation-butterfly, which both
    # have title "Butterfly", correspond to butterfly.svg.
    # Incomplete pages have page.path = "" so won't be linked to here.
    svg_name = page.title.lower().replace(" ", "-")
    return vm.Card(
        text=f"""
           ![](assets/images/charts/{svg_name}.svg#chart-icon)

           #### {page.title}
           """,
        href=page.path,
    )


def make_homepage_container(chart_group: ChartGroup) -> vm.Container:
    """Makes a container with cards for each completed and incomplete chart in chart_group.

    Args:
        chart_group: group of charts to make container for.

    Returns: container with cards for each chart in chart_group.

    """
    # Pages are sorted in title's alphabetical order and deduplicated so that e.g. pages distribution-butterfly and
    # deviation-butterfly, which both have title "Butterfly", correspond to a single card.
    return vm.Container(
        title=chart_group.name,
        layout=vm.Layout(grid=[[0, 1, 1]], col_gap="40px"),
        components=[
            Markdown(text=chart_group.intro_text, classname="intro-text"),
            FlexContainer(
                components=[
                    make_chart_card(page)
                    for page in sorted(
                        _remove_duplicates(chart_group.pages + chart_group.incomplete_pages),
                        key=lambda page: page.title,
                    )
                ],
            ),
        ],
    )


def _remove_duplicates(pages: List[Union[vm.Page, IncompletePage]]) -> List[Union[vm.Page, IncompletePage]]:
    # Deduplicate pages that have the same title. Using reversed means that the page that is kept is the first one
    # in the dashboard. This will be the one that the card on the homepage links to.
    return list({page.title: page for page in reversed(pages)}.values())


def make_navlink(chart_group: ChartGroup) -> vm.NavLink:
    """Makes a navlink with icon and links to every complete page within chart_group.

    Args:
        chart_group: chart_group to make a navlink for.

    Returns: navlink for chart_group.

    """
    # Pages are sorted in alphabetical order within each chart group.
    return vm.NavLink(
        label=chart_group.name,
        pages={chart_group.name: [page.id for page in sorted(chart_group.pages, key=lambda page: page.title)]},
        icon=chart_group.icon,
    )


homepage = vm.Page(
    title="Overview",
    components=[
        vm.Tabs(tabs=[make_homepage_container(chart_group) for chart_group in [ALL_CHART_GROUP, *CHART_GROUPS]]),
    ],
)

# TODO: consider whether each chart group should have its own individual homepage,
# e.g. at http://localhost:8050/deviation/. This could just repeat the content of the tab from the homepage and would
# work nicely with the hierarchical navigation.
dashboard = vm.Dashboard(
    # ALL_CHART_GROUP.pages has duplicated pages, e.g. both distribution-butterfly and deviation-butterfly.
    title="Vizro - Visual Vocabulary",
    pages=[homepage, *ALL_CHART_GROUP.pages],
    navigation=vm.Navigation(
        nav_selector=vm.NavBar(
            items=[
                vm.NavLink(label="Overview", pages=[homepage.id], icon="Home"),
            ]
            + [make_navlink(chart_group) for chart_group in CHART_GROUPS]
        )
    ),
)

if __name__ == "__main__":
    Vizro().build(dashboard).run()
